import copy
import warnings
import weakref

import pandas as pd

from woodwork.accessor_utils import (
    _check_column_schema,
    _is_dataframe,
    _is_series,
    init_series,
    is_schema_valid
)
from woodwork.column_schema import ColumnSchema
from woodwork.exceptions import (
    ParametersIgnoredWarning,
    TypingInfoMismatchWarning
)
from woodwork.indexers import _iLocIndexer, _locIndexer
from woodwork.logical_types import LatLong, Ordinal
from woodwork.statistics_utils import _get_box_plot_info_for_column
from woodwork.table_schema import TableSchema
from woodwork.utils import (
    _get_column_logical_type,
    _is_valid_latlong_series,
    import_or_none
)

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


class WoodworkColumnAccessor:
    def __init__(self, series):
        self._series_weakref = weakref.ref(series)
        self._schema = None

    def init(self, logical_type=None, semantic_tags=None,
             use_standard_tags=True, description=None, origin=None,
             metadata=None, schema=None, validate=True):
        """Initializes Woodwork typing information for a Series.

        Args:
            logical_type (LogicalType or str, optional): The logical type that should be assigned
                to the series. If no value is provided, the LogicalType for the series will
                be inferred. If the LogicalType provided or inferred does not have a dtype that
                is compatible with the series dtype, an error will be raised.
            semantic_tags (str or list or set, optional): Semantic tags to assign to the series.
                Defaults to an empty set if not specified. There are two options for
                specifying the semantic tags:
                (str) If only one semantic tag is being set, a single string can be passed.
                (list or set) If multiple tags are being set, a list or set of strings can be passed.
            use_standard_tags (bool, optional): If True, will add standard semantic tags to the series
                based on the inferred or specified logical type of the series. Defaults to True.
            description (str, optional): Optional text describing the contents of the series.
            origin (str, optional): Optional text specifying origin of the column (i.e. "base" or "engineered").
            metadata (dict[str -> json serializable], optional): Metadata associated with the series.
            schema (Woodwork.ColumnSchema, optional): Typing information to use for the Series instead of performing inference.
                Any other arguments provided will be ignored. Note that any changes made to the schema object after
                initialization will propagate to the Series. Similarly, to avoid unintended typing information changes,
                the same schema object should not be shared between Series.
            validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """

        if schema is not None:
            if validate:
                _validate_schema(schema, self._series)

            extra_params = []
            if logical_type is not None:
                extra_params.append('logical_type')
            if semantic_tags is not None:
                extra_params.append('semantic_tags')
            if description is not None:
                extra_params.append('description')
            if origin is not None:
                extra_params.append('origin')
            if metadata is not None:
                extra_params.append('metadata')
            if not use_standard_tags:
                extra_params.append('use_standard_tags')
            if extra_params:
                warnings.warn("A schema was provided and the following parameters were ignored: " + ", ".join(extra_params), ParametersIgnoredWarning)

            self._schema = schema
        else:
            logical_type = _get_column_logical_type(self._series, logical_type, self._series.name)

            if validate:
                self._validate_logical_type(logical_type)

            self._schema = ColumnSchema(logical_type=logical_type,
                                        semantic_tags=semantic_tags,
                                        use_standard_tags=use_standard_tags,
                                        description=description,
                                        origin=origin,
                                        metadata=metadata,
                                        validate=validate)

    @property
    def _series(self):
        return self._series_weakref()

    @property
    def schema(self):
        return copy.deepcopy(self._schema)

    @property
    @_check_column_schema
    def description(self):
        """The description of the series"""
        return self._schema.description

    @description.setter
    @_check_column_schema
    def description(self, description):
        self._schema.description = description

    @property
    @_check_column_schema
    def origin(self):
        """The origin of the series"""
        return self._schema.origin

    @origin.setter
    @_check_column_schema
    def origin(self, origin):
        self._schema.origin = origin

    @property
    @_check_column_schema
    def iloc(self):
        """
        Integer-location based indexing for selection by position.
        ``.iloc[]`` is primarily integer position based (from ``0`` to
        ``length-1`` of the axis), but may also be used with a boolean array.

        If the selection result is a Series, Woodwork typing information will
        be initialized for the returned Series.

        Allowed inputs are:
            An integer, e.g. ``5``.
            A list or array of integers, e.g. ``[4, 3, 0]``.
            A slice object with ints, e.g. ``1:7``.
            A boolean array.
            A ``callable`` function with one argument (the calling Series, DataFrame
            or Panel) and that returns valid output for indexing (one of the above).
            This is useful in method chains, when you don't have a reference to the
            calling object, but would like to base your selection on some value.
        """
        return _iLocIndexer(self._series)

    @property
    @_check_column_schema
    def loc(self):
        """
        Access a group of rows by label(s) or a boolean array.

        ``.loc[]`` is primarily label based, but may also be used with a
        boolean array.

        If the selection result is a Series, Woodwork typing information will
        be initialized for the returned Series.

        Allowed inputs are:
            A single label, e.g. ``5`` or ``'a'``, (note that ``5`` is
            interpreted as a *label* of the index, and **never** as an
            integer position along the index).
            A list or array of labels, e.g. ``['a', 'b', 'c']``.
            A slice object with labels, e.g. ``'a':'f'``.
            A boolean array of the same length as the axis being sliced,
            e.g. ``[True, False, True]``.
            An alignable boolean Series. The index of the key will be aligned before
            masking.
            An alignable Index. The Index of the returned selection will be the input.
            A ``callable`` function with one argument (the calling Series or
            DataFrame) and that returns valid output for indexing (one of the above)
        """
        return _locIndexer(self._series)

    @property
    @_check_column_schema
    def logical_type(self):
        """The logical type of the series"""
        return self._schema.logical_type

    @property
    @_check_column_schema
    def metadata(self):
        """The metadata of the series"""
        return self._schema.metadata

    @metadata.setter
    @_check_column_schema
    def metadata(self, metadata):
        self._schema.metadata = metadata

    @property
    @_check_column_schema
    def semantic_tags(self):
        """The semantic tags assigned to the series"""
        return self._schema.semantic_tags

    @property
    @_check_column_schema
    def use_standard_tags(self):
        return self._schema.use_standard_tags

    def __eq__(self, other, deep=True):
        if not self._schema.__eq__(other._schema, deep=deep):
            return False
        if self._series.name != other._series.name:
            return False
        if deep and isinstance(self._series, pd.Series):
            return self._series.equals(other._series)
        return True

    @_check_column_schema
    def __getattr__(self, attr):
        # Called if method is not present on the Accessor
        # If the method is present on Series, uses that method.
        if hasattr(self._series, attr):
            return self._make_series_call(attr)
        else:
            raise AttributeError(f"Woodwork has no attribute '{attr}'")

    @_check_column_schema
    def __repr__(self):
        msg = u"<Series: {} ".format(self._series.name)
        msg += u"(Physical Type = {}) ".format(self._series.dtype)
        msg += u"(Logical Type = {}) ".format(self.logical_type)
        msg += u"(Semantic Tags = {})>".format(self.semantic_tags)
        return msg

    def _make_series_call(self, attr):
        """Forwards the requested attribute onto the series object. Intercepts return value,
        attempting to initialize Woodwork with the current schema when a new Series is returned.
        Confirms schema is still valid for the original Series."""
        series_attr = getattr(self._series, attr)

        if callable(series_attr):
            def wrapper(*args, **kwargs):
                # Make Series call and intercept the result
                result = series_attr(*args, **kwargs)

                # Try to initialize Woodwork with the existing schema
                if _is_series(result):
                    valid_dtype = self._schema.logical_type._get_valid_dtype(type(result))
                    if str(result.dtype) == valid_dtype:
                        result.ww.init(schema=self.schema, validate=False)
                    else:
                        invalid_schema_message = 'dtype mismatch between original dtype, ' \
                            f'{valid_dtype}, and returned dtype, {result.dtype}'
                        warning_message = TypingInfoMismatchWarning().get_warning_message(attr,
                                                                                          invalid_schema_message,
                                                                                          'Series')
                        warnings.warn(warning_message, TypingInfoMismatchWarning)
                elif _is_dataframe(result):
                    # Initialize Woodwork if a valid table schema can be created from the original column schema
                    col_schema = self.schema
                    table_schema = TableSchema(column_names=[self.name],
                                               logical_types={self.name: col_schema.logical_type},
                                               semantic_tags={self.name: col_schema.semantic_tags},
                                               column_metadata={self.name: col_schema.metadata},
                                               use_standard_tags={self.name: col_schema.use_standard_tags},
                                               column_descriptions={self.name: col_schema.description},
                                               column_origins={self.name: col_schema.origin},
                                               validate=False)
                    if is_schema_valid(result, table_schema):
                        result.ww.init(schema=table_schema)
                # Always return the results of the Series operation whether or not Woodwork is initialized
                return result
            return wrapper
        # Directly return non-callable Series attributes
        return series_attr

    def _validate_logical_type(self, logical_type):
        """Validates that a logical type is consistent with the series dtype. Performs additional type
        specific validation, as required."""
        valid_dtype = logical_type._get_valid_dtype(type(self._series))
        if valid_dtype != str(self._series.dtype):
            raise ValueError(f"Cannot initialize Woodwork. Series dtype '{self._series.dtype}' is "
                             f"incompatible with {logical_type} dtype. Try converting series "
                             f"dtype to '{valid_dtype}' before initializing or use the "
                             "woodwork.init_series function to initialize.")

        if isinstance(logical_type, Ordinal):
            logical_type._validate_data(self._series)
        elif isinstance(logical_type, LatLong):
            if not _is_valid_latlong_series(self._series):
                raise ValueError("Cannot initialize Woodwork. Series does not contain properly formatted "
                                 "LatLong data. Try reformatting before initializing or use the "
                                 "woodwork.init_series function to initialize.")

    @_check_column_schema
    def add_semantic_tags(self, semantic_tags):
        """Add the specified semantic tags to the set of tags.

        Args:
            semantic_tags (str/list/set): New semantic tag(s) to add
        """
        self._schema._add_semantic_tags(semantic_tags,
                                        self._series.name)

    @_check_column_schema
    def remove_semantic_tags(self, semantic_tags):
        """Removes specified semantic tags from the current tags.

        Args:
            semantic_tags (str/list/set): Semantic tag(s) to remove.
        """
        self._schema._remove_semantic_tags(semantic_tags,
                                           self._series.name)

    @_check_column_schema
    def reset_semantic_tags(self):
        """Reset the semantic tags to the default values. The default values
        will be either an empty set or a set of the standard tags based on the
        column logical type, controlled by the use_standard_tags property.

        Args:
            None
        """
        self._schema._reset_semantic_tags()

    @_check_column_schema
    def set_logical_type(self, logical_type):
        """Update the logical type for the series, clearing any previously set semantic tags,
        and returning a new series with Woodwork initialied.

        Args:
            logical_type (LogicalType, str): The new logical type to set for the series.

        Returns:
            Series: A new series with the updated logical type.
        """
        # Create a new series without a schema to prevent new series from sharing a common
        # schema with current series
        new_series = self._series.copy()
        new_series._schema = None
        return init_series(new_series,
                           logical_type=logical_type,
                           semantic_tags=None,
                           use_standard_tags=self._schema.use_standard_tags,
                           description=self.description,
                           origin=self.origin,
                           metadata=copy.deepcopy(self.metadata))

    @_check_column_schema
    def set_semantic_tags(self, semantic_tags):
        """Replace current semantic tags with new values. If `use_standard_tags` is set
        to True for the series, any standard tags associated with the LogicalType of the
        series will be added as well.

        Args:
            semantic_tags (str/list/set): New semantic tag(s) to set
        """
        self._schema._set_semantic_tags(semantic_tags)

    @_check_column_schema
    def box_plot_dict(self, quantiles=None):
        """Gets the information necessary to create a box and whisker plot with outliers for a numeric column
        using the IQR method.

        Args:
            quantiles (dict[float -> float], optional): A dictionary containing the quantiles for the data
                where the key indicates the quantile, and the value is the quantile's value for the data. If
                no qantiles are provided, they will be computed from the data.

        Note:
            The minimum quantiles necessary for outlier detection using the IQR method are the
            first quartile (0.25) and third quartile (0.75). If these keys are missing from the quantiles dictionary,
            the following quantiles will be calculated: {0.0, 0.25, 0.5, 0.75, 1.0}, which correspond to
            {min, first quantile, median, third quantile, max}.

        Returns:
            (dict[str -> float,list[number]]): Returns a dictionary containing box plot information for the Series.
                The following elements will be found in the dictionary:

                - low_bound (float): the lower bound below which outliers lay - to be used as a whisker
                - high_bound (float): the high bound above which outliers lay - to be used as a whisker
                - quantiles (list[float]): the quantiles used to determine the bounds.
                    If quantiles were passed in, will contain all quantiles passed in. Otherwise, contains the five
                    quantiles {0.0, 0.25, 0.5, 0.75, 1.0}.
                - low_values (list[float, int]): the values of the lower outliers
                - high_values (list[float, int]): the values of the upper outliers
                - low_indices (list[int]): the corresponding index values for each of the lower outliers
                - high_indices (list[int]): the corresponding index values for each of the upper outliers
        """
        return _get_box_plot_info_for_column(self._series, quantiles=quantiles)


def _validate_schema(schema, series):
    if not isinstance(schema, ColumnSchema):
        raise TypeError('Provided schema must be a Woodwork.ColumnSchema object.')

    valid_dtype = schema.logical_type._get_valid_dtype(type(series))
    if str(series.dtype) != valid_dtype:
        raise ValueError(f"dtype mismatch between Series dtype {series.dtype}, and {schema.logical_type} dtype, {valid_dtype}")


@pd.api.extensions.register_series_accessor('ww')
class PandasColumnAccessor(WoodworkColumnAccessor):
    pass


if dd:
    @dd.extensions.register_series_accessor('ww')
    class DaskColumnAccessor(WoodworkColumnAccessor):
        pass


if ks:
    from databricks.koalas.extensions import register_series_accessor

    @register_series_accessor('ww')
    class KoalasColumnAccessor(WoodworkColumnAccessor):
        pass



import copy

from woodwork.accessor_utils import (
    _is_dask_dataframe,
    _is_dask_series,
    _is_dataframe,
    _is_series
)


class _iLocIndexer:
    def __init__(self, data):
        self.data = data
        if _is_dask_dataframe(data):
            raise TypeError("iloc is not supported for Dask DataFrames")
        elif _is_dask_series(data):
            raise TypeError("iloc is not supported for Dask Series")

    def __getitem__(self, key):
        selection = self.data.iloc[key]
        return _process_selection(selection, self.data)


class _locIndexer:
    def __init__(self, data):
        self.data = data

    def __getitem__(self, key):
        selection = self.data.loc[key]
        return _process_selection(selection, self.data)


def _process_selection(selection, original_data):
    if _is_series(selection):
        if _is_dask_series(selection):
            # Dask index values are a delayed object - can't compare below without computing
            index_vals = selection.index.values.compute()
        else:
            index_vals = selection.index.values
        if _is_dataframe(original_data) and set(index_vals) == set(original_data.columns):
            # Selecting a single row from a DataFrame, returned as Series without Woodwork initialized
            schema = None
        elif _is_dataframe(original_data):
            # Selecting a single column from a DataFrame
            schema = original_data.ww.schema.columns[selection.name]
        else:
            # Selecting a new Series from an existing Series
            schema = original_data.ww._schema
        if schema:
            selection.ww.init(schema=copy.deepcopy(schema), validate=False)
    elif _is_dataframe(selection):
        # Selecting a new DataFrame from an existing DataFrame
        schema = original_data.ww.schema
        new_schema = schema._get_subset_schema(list(selection.columns))
        selection.ww.init(schema=new_schema, validate=False)
    # Selecting a single value or return selection from above
    return selection



import inspect
import re

import pandas as pd
import pytest
from mock import patch

from woodwork import type_system
from woodwork.exceptions import ColumnNotPresentError, DuplicateTagsWarning
from woodwork.logical_types import (
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    PersonFullName,
    PhoneNumber,
    Unknown
)
from woodwork.table_schema import TableSchema


def test_schema_logical_types(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    assert isinstance(schema.logical_types, dict)
    assert set(schema.logical_types.keys()) == set(sample_column_names)
    for k, v in schema.logical_types.items():
        assert v == schema.columns[k].logical_type


def test_schema_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'age': ['numeric', 'age']
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags=semantic_tags)
    assert isinstance(schema.semantic_tags, dict)
    assert set(schema.semantic_tags.keys()) == set(sample_column_names)
    for k, v in schema.semantic_tags.items():
        assert isinstance(v, set)
        assert v == schema.columns[k].semantic_tags


def test_schema_types(sample_column_names, sample_inferred_logical_types):
    sample_column_names.append('formatted_date')

    ymd_format = Datetime(datetime_format='%Y~%m~%d')
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, 'formatted_date': ymd_format},
        use_standard_tags=True
    )

    returned_types = schema.types
    assert isinstance(returned_types, pd.DataFrame)
    assert 'Logical Type' in returned_types.columns
    assert 'Semantic Tag(s)' in returned_types.columns
    assert returned_types.shape[1] == 2
    assert len(returned_types.index) == len(sample_column_names)
    correct_logical_types = {name: ltype() for name, ltype in sample_inferred_logical_types.items()}
    correct_logical_types['formatted_date'] = ymd_format
    correct_logical_types = pd.Series(list(correct_logical_types.values()),
                                      index=list(correct_logical_types.keys()))
    assert correct_logical_types.equals(returned_types['Logical Type'])

    correct_semantic_tags = {'id': "['numeric']",
                             'full_name': '[]',
                             'email': '[]',
                             'phone_number': '[]',
                             'age': "['numeric']",
                             'signup_date': '[]',
                             'is_registered': '[]',
                             'double': "['numeric']",
                             'double_with_nan': "['numeric']",
                             'integer': "['numeric']",
                             'nullable_integer': "['numeric']",
                             'boolean': '[]',
                             'categorical': "['category']",
                             'datetime_with_NaT': '[]',
                             'formatted_date': '[]'}
    correct_semantic_tags = pd.Series(list(correct_semantic_tags.values()),
                                      index=list(correct_semantic_tags.keys()))
    assert correct_semantic_tags.equals(returned_types['Semantic Tag(s)'])


def test_schema_repr(small_df):
    schema = TableSchema(list(small_df.columns), logical_types={'sample_datetime_series': Datetime})

    schema_repr = repr(schema)
    expected_repr = '                       Logical Type Semantic Tag(s)\nColumn                                             \nsample_datetime_series     Datetime              []'
    assert schema_repr == expected_repr

    schema_html_repr = schema._repr_html_()
    expected_repr = '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>Logical Type</th>\n      <th>Semantic Tag(s)</th>\n    </tr>\n    <tr>\n      <th>Column</th>\n      <th></th>\n      <th></th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr>\n      <th>sample_datetime_series</th>\n      <td>Datetime</td>\n      <td>[]</td>\n    </tr>\n  </tbody>\n</table>'
    assert schema_html_repr == expected_repr


def test_schema_repr_empty():
    schema = TableSchema([], {})
    assert repr(schema) == 'Empty DataFrame\nColumns: [Logical Type, Semantic Tag(s)]\nIndex: []'

    assert schema._repr_html_() == '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>Logical Type</th>\n      <th>Semantic Tag(s)</th>\n    </tr>\n    <tr>\n      <th>Column</th>\n      <th></th>\n      <th></th>\n    </tr>\n  </thead>\n  <tbody>\n  </tbody>\n</table>'


def test_schema_equality(sample_column_names, sample_inferred_logical_types):
    schema_basic = TableSchema(sample_column_names, sample_inferred_logical_types)
    schema_basic2 = TableSchema(sample_column_names, sample_inferred_logical_types)
    schema_names = TableSchema(sample_column_names, sample_inferred_logical_types, name='test')

    assert schema_basic != schema_names
    assert schema_basic == schema_basic2

    missing_col_names = sample_column_names[1:]
    missing_logical_types = sample_inferred_logical_types.copy()
    missing_logical_types.pop('id')

    schema_missing_col = TableSchema(missing_col_names, missing_logical_types)
    assert schema_basic != schema_missing_col

    schema_index = TableSchema(sample_column_names, sample_inferred_logical_types, index='id')
    schema_time_index = TableSchema(sample_column_names, sample_inferred_logical_types, time_index='signup_date')

    assert schema_basic != schema_index
    assert schema_index != schema_time_index

    schema_numeric_time_index = TableSchema(sample_column_names, sample_inferred_logical_types, time_index='id')

    assert schema_time_index != schema_numeric_time_index

    schema_with_ltypes = TableSchema(sample_column_names,
                                     logical_types={**sample_inferred_logical_types, 'full_name': Categorical},
                                     time_index='signup_date')
    assert schema_with_ltypes != schema_time_index

    schema_with_metadata = TableSchema(sample_column_names, sample_inferred_logical_types, index='id', table_metadata={'created_by': 'user1'})
    assert TableSchema(sample_column_names, sample_inferred_logical_types, index='id') != schema_with_metadata
    assert TableSchema(sample_column_names,
                       sample_inferred_logical_types,
                       index='id',
                       table_metadata={'created_by': 'user1'}) == schema_with_metadata
    assert TableSchema(sample_column_names,
                       sample_inferred_logical_types,
                       index='id',
                       table_metadata={'created_by': 'user2'}) != schema_with_metadata


def test_schema_equality_standard_tags(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, use_standard_tags=True)
    no_standard_tags_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                          use_standard_tags=False)

    # Different standard tags values - using logical types with standard tags
    assert schema.use_standard_tags != no_standard_tags_schema.use_standard_tags
    assert schema != no_standard_tags_schema

    # Different standard tags values - no logical types that have standard tags
    unknown_types = {col_name: Unknown for col_name in sample_column_names}
    schema.set_types(logical_types=unknown_types)
    no_standard_tags_schema.set_types(logical_types=unknown_types)

    assert schema.use_standard_tags != no_standard_tags_schema.use_standard_tags
    assert schema != no_standard_tags_schema


def test_schema_shallow_equality(sample_column_names, sample_inferred_logical_types):
    metadata_table_1 = TableSchema(sample_column_names, sample_inferred_logical_types,
                                   table_metadata={'user': 'user0'}
                                   )
    metadata_table_2 = TableSchema(sample_column_names, sample_inferred_logical_types,
                                   table_metadata={'user': 'user0'}
                                   )

    assert metadata_table_1.__eq__(metadata_table_2, deep=False)
    assert metadata_table_1.__eq__(metadata_table_2, deep=True)

    diff_metadata_table = TableSchema(sample_column_names, sample_inferred_logical_types,
                                      table_metadata={'user': 'user1'}
                                      )

    assert metadata_table_1.__eq__(diff_metadata_table, deep=False)
    assert not metadata_table_1.__eq__(diff_metadata_table, deep=True)

    diff_col_metadata_table = TableSchema(sample_column_names, sample_inferred_logical_types,
                                          table_metadata={'user': 'user0'},
                                          column_metadata={'id': {'interesting_values': [0]}}
                                          )

    assert metadata_table_1.__eq__(diff_col_metadata_table, deep=False)
    assert not metadata_table_1.__eq__(diff_col_metadata_table, deep=True)

    diff_ltype_table = TableSchema(sample_column_names, {**sample_inferred_logical_types, 'id': Categorical},
                                   table_metadata={'user': 'user0'}
                                   )
    assert not metadata_table_1.__eq__(diff_ltype_table, deep=False)
    assert not metadata_table_1.__eq__(diff_ltype_table, deep=True)


def test_schema_table_metadata(sample_column_names, sample_inferred_logical_types):
    metadata = {'secondary_time_index': {'is_registered': 'age'}, 'date_created': '11/13/20'}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    assert schema.metadata == {}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         table_metadata=metadata, time_index='signup_date')
    assert schema.metadata == metadata


def test_column_schema_metadata(sample_column_names, sample_inferred_logical_types):
    column_metadata = {'metadata_field': [1, 2, 3], 'created_by': 'user0'}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    assert schema.columns['id'].metadata == {}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types, column_metadata={'id': column_metadata})
    assert schema.columns['id'].metadata == column_metadata


def test_filter_schema_cols_include(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date',
                         index='id',
                         name='df_name',
                         use_standard_tags=True)

    filtered = schema._filter_cols(include=Datetime)
    expected = {'signup_date', 'datetime_with_NaT'}
    assert set(filtered) == expected

    filtered = schema._filter_cols(include='email', col_names=True)
    assert filtered == ['email']

    filtered_log_type_string = schema._filter_cols(include='Unknown')
    filtered_log_type = schema._filter_cols(include=Unknown)
    expected = {'full_name', 'phone_number'}
    assert filtered_log_type == filtered_log_type_string
    assert set(filtered_log_type) == expected
    expected = {'integer', 'double', 'double_with_nan', 'age', 'nullable_integer'}
    filtered_semantic_tag = schema._filter_cols(include='numeric')
    assert set(filtered_semantic_tag) == expected

    filtered_multiple_overlap = schema._filter_cols(include=['Unknown', 'email'], col_names=True)
    expected = ['full_name', 'phone_number', 'email']
    for col in filtered_multiple_overlap:
        assert col in expected


def test_filter_schema_cols_exclude(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date',
                         index='id',
                         name='df_name',
                         use_standard_tags=True)

    filtered = schema._filter_cols(exclude=Datetime)
    assert 'signup_date' not in filtered

    filtered = schema._filter_cols(exclude='email', col_names=True)
    assert 'email' not in filtered

    filtered_log_type_string = schema._filter_cols(exclude='Unknown')
    filtered_log_type = schema._filter_cols(exclude=Unknown)
    expected = {'boolean',
                'double',
                'datetime_with_NaT',
                'email',
                'categorical',
                'double_with_nan',
                'signup_date',
                'age',
                'integer',
                'nullable_integer',
                'is_registered',
                'id'}
    assert filtered_log_type == filtered_log_type_string
    assert set(filtered_log_type) == expected

    filtered_semantic_tag = schema._filter_cols(exclude='numeric')
    assert 'age' not in filtered_semantic_tag

    filtered_multiple_overlap = schema._filter_cols(exclude=['Unknown', 'email'], col_names=True)
    expected = {'is_registered',
                'double',
                'nullable_integer',
                'double_with_nan',
                'categorical',
                'datetime_with_NaT', 'age',
                'integer',
                'signup_date',
                'boolean',
                'id'}
    assert set(filtered_multiple_overlap) == expected


def test_filter_schema_cols_no_matches(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date',
                         index='id',
                         name='df_name')

    filter_no_matches = schema._filter_cols(include='nothing')
    assert filter_no_matches == []

    filter_empty_list = schema._filter_cols(include=[])
    assert filter_empty_list == []

    filter_non_string = schema._filter_cols(include=1)
    assert filter_non_string == []

    filter_exclude_no_matches = schema._filter_cols(exclude='nothing')
    assert set(filter_exclude_no_matches) == set(sample_column_names)

    filter_exclude_empty_list = schema._filter_cols(exclude=[])
    assert set(filter_exclude_empty_list) == set(sample_column_names)

    filter_exclude_non_string = schema._filter_cols(exclude=1)
    assert set(filter_exclude_non_string) == set(sample_column_names)


def test_filter_schema_errors(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date',
                         index='id',
                         name='df_name')

    err_msg = "Invalid selector used in include: {} must be a string, uninstantiated and registered LogicalType, or valid column name"
    with pytest.raises(TypeError, match=err_msg):
        schema._filter_cols(include=['boolean', 'index', Double, {}])

    err_msg = "Invalid selector used in include: {} must be a string, uninstantiated and registered LogicalType, or valid column name"
    with pytest.raises(TypeError, match=err_msg):
        schema._filter_cols(include=['boolean', 'index', Double, {}], col_names=True)

    err_msg = "Invalid selector used in include: Datetime cannot be instantiated"
    with pytest.raises(TypeError, match=err_msg):
        schema._filter_cols(Datetime())

    type_system.remove_type(EmailAddress)
    err_msg = "Specified LogicalType selector EmailAddress is not registered in Woodwork's type system."
    with pytest.raises(TypeError, match=err_msg):
        schema._filter_cols(EmailAddress)

    err_msg = "Invalid selector used in include: EmailAddress must be a string, uninstantiated and registered LogicalType, or valid column name"
    with pytest.raises(TypeError, match=err_msg):
        schema._filter_cols(EmailAddress())
    type_system.reset_defaults()


def test_filter_schema_overlap_name_and_type(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    filter_name_ltype_overlap = schema._filter_cols(include='full_name')
    assert filter_name_ltype_overlap == []

    filter_overlap_with_name = schema._filter_cols(include='full_name', col_names=True)
    assert filter_overlap_with_name == ['full_name']

    schema = TableSchema(sample_column_names,
                         {**sample_inferred_logical_types, 'full_name': Categorical, 'age': PersonFullName},
                         semantic_tags={'id': 'person_full_name'})

    filter_tag_and_ltype = schema._filter_cols(include='person_full_name')
    assert set(filter_tag_and_ltype) == {'id', 'age'}

    filter_all_three = schema._filter_cols(include=['person_full_name', 'full_name'], col_names=True)
    assert set(filter_all_three) == {'id', 'age', 'full_name'}


def test_filter_schema_non_string_cols():
    schema = TableSchema(
        column_names=[0, 1, 2, 3],
        logical_types={0: Integer, 1: Categorical, 2: Unknown, 3: Double},
        use_standard_tags=True
    )

    filter_types_and_tags = schema._filter_cols(include=[Integer, 'category'])
    assert filter_types_and_tags == [0, 1]

    filter_by_name = schema._filter_cols(include=[0, 1], col_names=True)
    assert filter_by_name == [0, 1]


def test_get_subset_schema(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    new_schema = schema._get_subset_schema(sample_column_names[1:4])
    for col in new_schema.columns:
        assert new_schema.semantic_tags[col] == schema.semantic_tags[col]
        assert new_schema.logical_types[col] == schema.logical_types[col]


def test_get_subset_schema_all_params(sample_column_names, sample_inferred_logical_types):
    # The first element is self, so it won't be included in kwargs
    possible_schema_params = inspect.getfullargspec(TableSchema.__init__)[0][1:]

    kwargs = {
        'column_names': sample_column_names,
        'logical_types': {**sample_inferred_logical_types, 'email': EmailAddress},
        'name': 'test_dt',
        'index': 'id',
        'time_index': 'signup_date',
        'semantic_tags': {'age': 'test_tag'},
        'table_metadata': {'created_by': 'user1'},
        'column_metadata': {'phone_number': {'format': 'xxx-xxx-xxxx'}},
        'use_standard_tags': False,
        'column_descriptions': {'age': 'this is a description'},
        'column_origins': 'base',
        'validate': True
    }

    # Confirm all possible params to TableSchema init are present with non-default values where possible
    assert set(possible_schema_params) == set(kwargs.keys())

    schema = TableSchema(**kwargs)
    copy_schema = schema._get_subset_schema(sample_column_names)

    assert schema == copy_schema
    assert schema is not copy_schema


def test_set_logical_types(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'phone_number': ['tag3', 'tag2'],
        'signup_date': {'secondary_time_index'},
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=True)

    schema.set_types(logical_types={
        'full_name': Categorical,
        'email': EmailAddress,
        'phone_number': PhoneNumber,
        'age': Double,
    })

    assert schema.logical_types['full_name'] == Categorical
    assert schema.logical_types['email'] == EmailAddress
    assert schema.logical_types['phone_number'] == PhoneNumber
    assert schema.logical_types['age'] == Double

    # Verify semantic tags were reset to standard tags in columns with Logical Type changes
    assert schema.semantic_tags['full_name'] == {'category'}
    assert schema.semantic_tags['email'] == set()
    assert schema.semantic_tags['phone_number'] == set()
    assert schema.semantic_tags['age'] == {'numeric'}

    # Verify signup date column was unchanged
    assert isinstance(schema.logical_types['signup_date'], Datetime)
    assert schema.semantic_tags['signup_date'] == {'secondary_time_index'}


def test_set_logical_types_empty(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'age': 'test_tag'
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='full_name',
                         semantic_tags=semantic_tags, use_standard_tags=True)

    # An empty set should reset the tags
    schema.set_types(semantic_tags={'full_name': set()}, retain_index_tags=False)
    assert isinstance(schema.logical_types['full_name'], Unknown)
    assert schema.semantic_tags['full_name'] == set()

    schema.set_types(semantic_tags={'age': set()})
    assert isinstance(schema.logical_types['age'], IntegerNullable)
    assert schema.semantic_tags['age'] == {'numeric'}


def test_set_logical_types_invalid_data(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    error_message = re.escape("logical_types contains columns that are not present in TableSchema: ['birthday']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        schema.set_types(logical_types={'birthday': Double})

    error_message = ("Logical Types must be of the LogicalType class "
                     "and registered in Woodwork's type system. "
                     "Double does not meet that criteria.")
    with pytest.raises(TypeError, match=error_message):
        schema.set_types(logical_types={'id': 'Double'})

    error_message = ("Logical Types must be of the LogicalType class "
                     "and registered in Woodwork's type system. "
                     "<class 'int'> does not meet that criteria.")
    with pytest.raises(TypeError, match=error_message):
        schema.set_types(logical_types={'age': int})

    error_message = "semantic_tags for full_name must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        schema.set_types(semantic_tags={'full_name': None})


def test_set_types_combined(sample_column_names, sample_inferred_logical_types):
    # test that the resetting of indices when ltype changes doesnt touch index??
    semantic_tags = {
        'id': 'tag1',
        'email': ['tag2'],
        'phone_number': ['tag3', 'tag2'],
        'signup_date': {'secondary_time_index'},
    }
    # use standard tags and keep index tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=True,
                         index='id')

    schema.set_types(semantic_tags={'id': 'new_tag', 'age': 'new_tag', 'email': 'new_tag'},
                     logical_types={'id': Double, 'age': Double, 'email': Categorical})

    assert schema.semantic_tags['id'] == {'new_tag', 'index'}
    assert schema.semantic_tags['age'] == {'numeric', 'new_tag'}
    assert schema.semantic_tags['email'] == {'new_tag', 'category'}

    assert schema.logical_types['id'] == Double
    assert schema.logical_types['age'] == Double
    assert schema.logical_types['email'] == Categorical

    # use standard tags and lose index tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=True,
                         index='id', time_index='age')

    schema.set_types(semantic_tags={'age': 'new_tag'},
                     logical_types={'id': Double},
                     retain_index_tags=False)

    assert schema.semantic_tags['id'] == {'numeric'}
    assert schema.semantic_tags['age'] == {'numeric', 'new_tag'}

    # don't use standard tags and lose index tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=False,
                         index='id', time_index='signup_date')

    schema.set_types(semantic_tags={'id': 'new_tag', 'age': 'new_tag', 'is_registered': 'new_tag'},
                     logical_types={'id': Double, 'age': Double, 'email': Categorical}, retain_index_tags=False)

    assert schema.index is None
    assert schema.time_index == 'signup_date'

    assert schema.semantic_tags['id'] == {'new_tag'}
    assert schema.semantic_tags['age'] == {'new_tag'}
    assert schema.semantic_tags['is_registered'] == {'new_tag'}
    assert schema.semantic_tags['email'] == set()
    assert schema.semantic_tags['signup_date'] == {'time_index', 'secondary_time_index'}

    assert schema.logical_types['id'] == Double
    assert schema.logical_types['age'] == Double
    assert schema.logical_types['email'] == Categorical


def test_set_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'age': ['numeric', 'age']
    }
    expected_tags = {
        'full_name': {'tag1'},
        'age': {'numeric', 'age'}
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags=semantic_tags)
    assert schema.semantic_tags['full_name'] == expected_tags['full_name']
    assert schema.semantic_tags['age'] == expected_tags['age']

    new_tags = {
        'full_name': ['new_tag'],
        'age': 'numeric',
    }
    schema.set_types(semantic_tags=new_tags)

    assert schema.semantic_tags['full_name'] == {'new_tag'}
    assert schema.semantic_tags['age'] == {'numeric'}


def test_set_semantic_tags_with_index(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=False)
    assert schema.semantic_tags['id'] == {'index'}

    new_tags = {
        'id': 'new_tag',
    }
    schema.set_types(semantic_tags=new_tags)
    assert schema.semantic_tags['id'] == {'index', 'new_tag'}

    schema.set_types(semantic_tags=new_tags, retain_index_tags=False)
    assert schema.semantic_tags['id'] == {'new_tag'}


def test_set_semantic_tags_with_time_index(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date', use_standard_tags=False)
    assert schema.semantic_tags['signup_date'] == {'time_index'}

    new_tags = {
        'signup_date': 'new_tag',
    }
    schema.set_types(semantic_tags=new_tags)
    assert schema.semantic_tags['signup_date'] == {'time_index', 'new_tag'}

    schema.set_types(semantic_tags=new_tags, retain_index_tags=False)
    assert schema.semantic_tags['signup_date'] == {'new_tag'}


def test_add_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'age': ['numeric', 'age']
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=False,
                         index='id')

    new_tags = {
        'full_name': ['list_tag'],
        'age': 'str_tag',
        'id': {'set_tag'}
    }
    schema.add_semantic_tags(new_tags)

    assert schema.semantic_tags['full_name'] == {'tag1', 'list_tag'}
    assert schema.semantic_tags['age'] == {'numeric', 'age', 'str_tag'}
    assert schema.semantic_tags['id'] == {'set_tag', 'index'}


def test_warns_on_adding_duplicate_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'age': ['numeric', 'age']
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         semantic_tags=semantic_tags, use_standard_tags=False,
                         index='id')

    new_tags = {
        'full_name': 'tag1',
    }
    expected_message = "Semantic tag(s) 'tag1' already present on column 'full_name'"
    with pytest.warns(DuplicateTagsWarning) as record:
        schema.add_semantic_tags(new_tags)
    assert len(record) == 1
    assert record[0].message.args[0] == expected_message

    assert schema.semantic_tags['full_name'] == {'tag1'}


def test_reset_all_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'age': 'age'
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags=semantic_tags, use_standard_tags=True)

    schema.reset_semantic_tags()
    assert schema.semantic_tags['full_name'] == set()
    assert schema.semantic_tags['age'] == {'numeric'}


def test_reset_semantic_tags_with_index(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'id': 'tag1',
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id',
                         semantic_tags=semantic_tags,
                         use_standard_tags=False)
    assert schema.semantic_tags['id'] == {'index', 'tag1'}

    schema.reset_semantic_tags('id', retain_index_tags=True)
    assert schema.semantic_tags['id'] == {'index'}

    schema.reset_semantic_tags('id')
    assert schema.semantic_tags['id'] == set()


def test_reset_semantic_tags_with_time_index(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'signup_date': 'tag1',
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         time_index='signup_date',
                         semantic_tags=semantic_tags,
                         use_standard_tags=False)
    assert schema.semantic_tags['signup_date'] == {'time_index', 'tag1'}

    schema.reset_semantic_tags('signup_date', retain_index_tags=True)
    assert schema.semantic_tags['signup_date'] == {'time_index'}

    schema.reset_semantic_tags('signup_date')
    assert schema.semantic_tags['signup_date'] == set()


def test_reset_semantic_tags_invalid_column(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,)
    error_msg = re.escape("Column(s) \'[\'invalid_column\']\' not found in DataFrame")
    with pytest.raises(ColumnNotPresentError, match=error_msg):
        schema.reset_semantic_tags('invalid_column')


def test_remove_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': ['tag1', 'tag2', 'tag3'],
        'age': ['numeric', 'age'],
        'id': ['tag1', 'tag2']
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags=semantic_tags, use_standard_tags=False)
    tags_to_remove = {
        'full_name': ['tag1', 'tag3'],
        'age': 'numeric',
        'id': {'tag1'}
    }
    schema.remove_semantic_tags(tags_to_remove)
    assert schema.semantic_tags['full_name'] == {'tag2'}
    assert schema.semantic_tags['age'] == {'age'}
    assert schema.semantic_tags['id'] == {'tag2'}


def test_raises_error_setting_index_tag_directly(sample_column_names, sample_inferred_logical_types):
    error_msg = re.escape("Cannot add 'index' tag directly for column id. To set a column as the index, "
                          "use DataFrame.ww.set_index() instead.")

    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    with pytest.raises(ValueError, match=error_msg):
        schema.add_semantic_tags({'id': 'index'})
    with pytest.raises(ValueError, match=error_msg):
        schema.set_types(semantic_tags={'id': 'index'})


def test_raises_error_setting_time_index_tag_directly(sample_column_names, sample_inferred_logical_types):
    error_msg = re.escape("Cannot add 'time_index' tag directly for column signup_date. To set a column as the time index, "
                          "use DataFrame.ww.set_time_index() instead.")
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    with pytest.raises(ValueError, match=error_msg):
        schema.add_semantic_tags({'signup_date': 'time_index'})
    with pytest.raises(ValueError, match=error_msg):
        schema.set_types(semantic_tags={'signup_date': 'time_index'})


def test_removes_index_via_tags(sample_column_names, sample_inferred_logical_types):
    # Check setting tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=True)
    schema.set_types(semantic_tags={'id': 'new_tag'}, retain_index_tags=False)
    assert schema.semantic_tags['id'] == {'numeric', 'new_tag'}
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=False)
    schema.set_types(semantic_tags={'id': 'new_tag'}, retain_index_tags=False)
    assert schema.semantic_tags['id'] == {'new_tag'}
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='full_name', use_standard_tags=True)
    schema.set_types(semantic_tags={'full_name': 'new_tag'}, retain_index_tags=False)
    assert schema.semantic_tags['full_name'] == {'new_tag'}
    assert schema.index is None

    # Check removing tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=True)
    schema.remove_semantic_tags(semantic_tags={'id': 'index'})
    assert schema.semantic_tags['id'] == {'numeric'}
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=False)
    schema.remove_semantic_tags(semantic_tags={'id': 'index'})
    assert schema.semantic_tags['id'] == set()
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='full_name', use_standard_tags=True)
    schema.remove_semantic_tags(semantic_tags={'full_name': 'index'})
    assert schema.semantic_tags['full_name'] == set()
    assert schema.index is None

    # Check resetting tags
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=True)
    schema.reset_semantic_tags('id')
    assert schema.semantic_tags['id'] == {'numeric'}
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', use_standard_tags=False)
    schema.reset_semantic_tags('id')
    assert schema.semantic_tags['id'] == set()
    assert schema.index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='full_name', use_standard_tags=True)
    schema.reset_semantic_tags('full_name')
    assert schema.semantic_tags['full_name'] == set()
    assert schema.index is None


def test_removes_time_index_via_tags(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, time_index='signup_date')
    schema.set_types(semantic_tags={'signup_date': 'new_tag'}, retain_index_tags=False)
    assert schema.semantic_tags['signup_date'] == {'new_tag'}
    assert schema.time_index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types, time_index='signup_date')
    schema.remove_semantic_tags(semantic_tags={'signup_date': 'time_index'})
    assert schema.semantic_tags['signup_date'] == set()
    assert schema.time_index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types, time_index='signup_date')
    schema.reset_semantic_tags('signup_date')
    assert schema.semantic_tags['signup_date'] == set()
    assert schema.time_index is None


def test_set_index(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, use_standard_tags=True)
    assert schema.index is None
    assert schema.semantic_tags['id'] == {'numeric'}
    assert schema.semantic_tags['age'] == {'numeric'}

    schema.set_index('id')
    assert schema.index == 'id'
    assert schema.semantic_tags['id'] == {'index'}

    schema.set_index('age')
    assert schema.index == 'age'
    assert schema.semantic_tags['age'] == {'index'}
    assert schema.semantic_tags['id'] == {'numeric'}

    schema.set_index(None)
    assert schema.index is None
    assert schema.semantic_tags['age'] == {'numeric'}
    assert schema.semantic_tags['id'] == {'numeric'}


def test_set_index_errors(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    error = re.escape("Specified index column `testing` not found in TableSchema.")
    with pytest.raises(ColumnNotPresentError, match=error):
        schema.set_index('testing')

    match = '"id" is already set as the index. '
    match += 'An index cannot also be the time index.'
    with pytest.raises(ValueError, match=match):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            index='id',
            time_index='id',
        )

    schema.set_index('id')
    with pytest.raises(ValueError, match=match):
        schema.set_time_index('id')

    schema.set_time_index('signup_date')
    match = '"signup_date" is already set as the time index. '
    match += 'A time index cannot also be the index.'
    with pytest.raises(ValueError, match=match):
        schema.set_index('signup_date')


def test_set_time_index(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, use_standard_tags=True)
    assert schema.time_index is None
    assert schema.semantic_tags['age'] == {'numeric'}
    assert schema.semantic_tags['signup_date'] == set()

    schema.set_time_index('signup_date')
    assert schema.time_index == 'signup_date'
    assert schema.semantic_tags['signup_date'] == {'time_index'}

    schema.set_time_index('age')
    assert schema.time_index == 'age'
    assert schema.semantic_tags['age'] == {'numeric', 'time_index'}
    assert schema.semantic_tags['signup_date'] == set()

    schema.set_time_index(None)
    assert schema.index is None
    assert schema.semantic_tags['age'] == {'numeric'}
    assert schema.semantic_tags['signup_date'] == set()


def test_set_numeric_datetime_time_index(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, {**sample_inferred_logical_types, 'age': Datetime})

    assert isinstance(schema.logical_types['age'], Datetime)
    assert schema.semantic_tags['age'] == set()

    schema.set_time_index('age')
    assert schema.time_index == 'age'
    assert schema.semantic_tags['age'] == {'time_index'}

    schema.set_time_index(None)
    assert schema.time_index is None
    assert schema.semantic_tags['age'] == set()


def test_set_time_index_errors(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    error = re.escape("Specified time index column `testing` not found in TableSchema")
    with pytest.raises(ColumnNotPresentError, match=error):
        schema.set_time_index('testing')

    error = re.escape("Time index column must be a Datetime or numeric column.")
    with pytest.raises(TypeError, match=error):
        schema.set_time_index('email')


def test_set_index_twice(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    schema.set_index(None)
    assert schema.index is None

    schema.set_time_index(None)
    assert schema.time_index is None

    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id', time_index='signup_date')
    original_schema = schema._get_subset_schema(list(schema.columns.keys()))

    schema.set_index('id')
    assert schema.index == 'id'
    assert schema.semantic_tags['id'] == {'index'}
    assert schema == original_schema

    schema.set_time_index('signup_date')
    assert schema.time_index == 'signup_date'
    assert schema.semantic_tags['signup_date'] == {'time_index'}
    assert schema == original_schema


def test_schema_rename_errors(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, index='id', time_index='signup_date')

    error = "columns must be a dictionary"
    with pytest.raises(TypeError, match=error):
        schema.rename(1)

    error = 'New columns names must be unique from one another.'
    with pytest.raises(ValueError, match=error):
        schema.rename({'age': 'test', 'full_name': 'test'})

    error = 'Column to rename must be present. not_present cannot be found.'
    with pytest.raises(ColumnNotPresentError, match=error):
        schema.rename({'not_present': 'test'})

    error = 'The column email is already present. Please choose another name to rename age to or also rename age.'
    with pytest.raises(ValueError, match=error):
        schema.rename({'age': 'email'})


def test_schema_rename(sample_column_names, sample_inferred_logical_types):

    table_metadata = {'table_info': 'this is text'}
    id_description = 'the id of the row'
    id_origin = 'base'
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id',
                         time_index='signup_date',
                         table_metadata=table_metadata,
                         column_descriptions={'id': id_description},
                         column_origins={'id': id_origin})
    original_schema = schema._get_subset_schema(list(schema.columns.keys()))

    renamed_schema = schema.rename({'age': 'birthday'})

    # Confirm original schema hasn't changed
    assert schema == original_schema

    assert 'age' not in renamed_schema.columns
    assert 'birthday' in renamed_schema.columns

    # confirm that metadata and descriptions are there
    assert renamed_schema.metadata == table_metadata
    assert schema.columns['id'].description == id_description
    assert schema.columns['id'].origin == id_origin

    old_col = schema.columns['age']
    new_col = renamed_schema.columns['birthday']
    assert old_col.logical_type == new_col.logical_type
    assert old_col.semantic_tags == new_col.semantic_tags

    swapped_schema = schema.rename({'age': 'full_name', 'full_name': 'age'})
    swapped_back_schema = swapped_schema.rename({'age': 'full_name', 'full_name': 'age'})
    assert swapped_back_schema == schema


def test_schema_rename_indices(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         index='id',
                         time_index='signup_date')

    renamed_schema = schema.rename({'id': 'renamed_index', 'signup_date': 'renamed_time_index'})
    assert 'id' not in renamed_schema.columns
    assert 'signup_date' not in renamed_schema.columns
    assert 'renamed_index' in renamed_schema.columns
    assert 'renamed_time_index' in renamed_schema.columns

    assert renamed_schema.index == 'renamed_index'
    assert renamed_schema.time_index == 'renamed_time_index'


@patch("woodwork.table_schema._check_time_index")
@patch("woodwork.table_schema._check_index")
def test_validation_methods_called(mock_check_index, mock_check_time_index,
                                   sample_column_names, sample_inferred_logical_types):
    validation_schema = TableSchema(sample_column_names, sample_inferred_logical_types)
    no_validation_schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    assert not mock_check_index.called
    assert not mock_check_time_index.called

    no_validation_schema.set_index('id', validate=False)
    assert not mock_check_index.called

    validation_schema.set_index('id', validate=True)
    assert mock_check_index.called
    assert validation_schema == no_validation_schema

    validation_schema.set_time_index('signup_date', validate=False)
    assert not mock_check_time_index.called

    no_validation_schema.set_time_index('signup_date', validate=True)
    assert mock_check_time_index.called
    assert validation_schema == no_validation_schema



import numpy as np
import pandas as pd
import pytest

from woodwork.logical_types import (
    Boolean,
    BooleanNullable,
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    PersonFullName,
    PhoneNumber,
    Unknown
)
from woodwork.utils import import_or_none


def pd_to_dask(series):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(series, npartitions=2)


def pd_to_koalas(series):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(series)


@pytest.fixture(scope='session', autouse=True)
def spark_session():
    pyspark = import_or_none('pyspark.sql')

    if pyspark:
        spark = pyspark.SparkSession.builder \
            .master("local[*]") \
            .config("spark.driver.extraJavaOptions", "-Dio.netty.tryReflectionSetAccessible=True") \
            .config("spark.sql.shuffle.partitions", "2") \
            .config("spark.driver.bindAddress", "127.0.0.1") \
            .getOrCreate()

        return spark


@pytest.fixture(params=['sample_df_pandas', 'sample_df_dask', 'sample_df_koalas'])
def sample_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_df_pandas():
    return pd.DataFrame({
        'id': range(4),
        'full_name': ['Mr. John Doe', 'Doe, Mrs. Jane', 'James Brown', 'Ms. Paige Turner'],
        'email': ['john.smith@example.com', np.nan, 'team@featuretools.com', 'junk@example.com'],
        'phone_number': ['5555555555', '555-555-5555', '1-(555)-555-5555', '555-555-5555'],
        'age': pd.Series([pd.NA, 33, 33, 57], dtype='Int64'),
        'signup_date': [pd.to_datetime('2020-09-01')] * 4,
        'is_registered': pd.Series([pd.NA, False, True, True], dtype='boolean'),
        'double': [0, 1, 0.0001, -5.4321],
        'double_with_nan': [np.nan, -123.45, 0.0001, 1],
        'integer': [-1, 4, 9, 25],
        'nullable_integer': pd.Series([pd.NA, -36, 49, 64], dtype='Int64'),
        'boolean': [True, False, False, True],
        'categorical': pd.Series(["a", "b", "c", "a"], dtype="category"),
        'datetime_with_NaT': [pd.to_datetime('2020-09-01')] * 3 + [pd.NaT],
    })


@pytest.fixture()
def sample_df_dask(sample_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(sample_df_pandas, npartitions=2)


@pytest.fixture()
def sample_df_koalas(sample_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(sample_df_pandas)


@pytest.fixture(params=['sample_unsorted_df_pandas', 'sample_unsorted_df_dask', 'sample_unsorted_df_koalas'])
def sample_unsorted_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_unsorted_df_pandas():
    return pd.DataFrame({
        'id': [3, 1, 2, 0],
        'full_name': ['Mr. John Doe', 'Doe, Mrs. Jane', 'James Brown', 'Ms. Paige Turner'],
        'email': ['john.smith@example.com', np.nan, 'team@featuretools.com', 'junk@example.com'],
        'phone_number': ['5555555555', '555-555-5555', '1-(555)-555-5555', '555-555-5555'],
        'age': [33, 25, 33, 57],
        'signup_date': pd.to_datetime(['2020-09-01', '2020-08-01', '2020-08-02', '2020-09-01']),
        'is_registered': [True, False, True, True],
    })


@pytest.fixture()
def sample_unsorted_df_dask(sample_unsorted_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason='Dask not installed, skipping')
    return dd.from_pandas(sample_unsorted_df_pandas, npartitions=2)


@pytest.fixture()
def sample_unsorted_df_koalas(sample_unsorted_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(sample_unsorted_df_pandas)


@pytest.fixture(params=['sample_series_pandas', 'sample_series_dask', 'sample_series_koalas'])
def sample_series(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_series_pandas():
    return pd.Series(['a', 'b', 'c'] + 10 * ['a', 'a', 'a'], name='sample_series').astype('category')


@pytest.fixture()
def sample_series_dask(sample_series_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(sample_series_pandas, npartitions=2)


@pytest.fixture()
def sample_series_koalas(sample_series_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(sample_series_pandas.astype('string'))


@pytest.fixture(params=['sample_datetime_series_pandas', 'sample_datetime_series_dask', 'sample_datetime_series_koalas'])
def sample_datetime_series(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_datetime_series_pandas():
    return pd.Series([pd.to_datetime('2020-09-01')] * 4, name='sample_datetime_series').astype('object')


@pytest.fixture()
def sample_datetime_series_dask(sample_datetime_series_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(sample_datetime_series_pandas, npartitions=2)


@pytest.fixture()
def sample_datetime_series_koalas(sample_datetime_series_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(sample_datetime_series_pandas)


@pytest.fixture()
def time_index_df_pandas():
    return pd.DataFrame({
        'id': [0, 1, 2, 3],
        'times': ['2019-01-01', '2019-01-02', '2019-01-03', pd.NA],
        'ints': [1, 2, 3, 4],
        'strs': ['1', '2', '3', '4'],
        'letters': ['a', 'b', 'c', 'd'],
        'bools': [True, False, False, True]
    })


@pytest.fixture()
def time_index_df_dask(time_index_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(time_index_df_pandas, npartitions=2)


@pytest.fixture()
def time_index_df_koalas(time_index_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(time_index_df_pandas)


@pytest.fixture(params=['time_index_df_pandas', 'time_index_df_dask', 'time_index_df_koalas'])
def time_index_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def numeric_time_index_df_pandas():
    return pd.DataFrame({
        'floats': pd.Series([1, 2, 3, 4], dtype='float'),
        'ints': pd.Series([1, -2, 3, 4], dtype='int64'),
        'with_null': pd.Series([1, pd.NA, 3, 4], dtype='Int64'),
    })


@pytest.fixture()
def numeric_time_index_df_dask(numeric_time_index_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(numeric_time_index_df_pandas, npartitions=2)


@pytest.fixture()
def numeric_time_index_df_koalas(numeric_time_index_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(numeric_time_index_df_pandas)


@pytest.fixture(params=['numeric_time_index_df_pandas', 'numeric_time_index_df_dask', 'numeric_time_index_df_koalas'])
def numeric_time_index_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def describe_df_pandas():
    index_data = [0, 1, 2, 3, 4, 5, 6, 7]
    boolean_data = [True, False, True, True, False, True, False, None]
    category_data = ['red', 'blue', 'red', np.nan, 'red', 'blue', 'red', 'yellow']
    datetime_data = pd.to_datetime(['2020-01-01',
                                    '2020-02-01',
                                    '2020-01-01 08:00',
                                    '2020-02-02 16:00',
                                    '2020-02-02 18:00',
                                    pd.NaT,
                                    '2020-02-01',
                                    '2020-01-02'])
    formatted_datetime_data = pd.Series(['2020~01~01',
                                         '2020~02~01',
                                         '2020~03~01',
                                         '2020~02~02',
                                         '2020~03~02',
                                         pd.NaT,
                                         '2020~02~01',
                                         '2020~01~02'])
    numeric_data = pd.Series([10, 20, 17, 32, np.nan, 1, 56, 10])
    natural_language_data = [
        'This is a natural language sentence',
        'Duplicate sentence.',
        'This line has numbers in it 000123.',
        'How about some symbols?!',
        'This entry contains two sentences. Second sentence.',
        'Duplicate sentence.',
        np.nan,
        'I am the last line',
    ]
    latlong_data = [(0, 0),
                    (1, 1),
                    (2, 2),
                    (3, 3),
                    (0, 0),
                    (np.nan, np.nan),
                    (np.nan, 6),
                    np.nan]
    timedelta_data = datetime_data - pd.Timestamp('2020-01-01')

    return pd.DataFrame({
        'index_col': index_data,
        'boolean_col': boolean_data,
        'category_col': category_data,
        'datetime_col': datetime_data,
        'formatted_datetime_col': formatted_datetime_data,
        'numeric_col': numeric_data,
        'natural_language_col': natural_language_data,
        'timedelta_col': timedelta_data,
        'latlong_col': latlong_data,
    })


@pytest.fixture()
def describe_df_dask(describe_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(describe_df_pandas, npartitions=2)


@pytest.fixture()
def describe_df_koalas(describe_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(describe_df_pandas
                          .applymap(lambda tup: [None if pd.isnull(elt) else elt for elt in tup] if isinstance(tup, tuple) else tup)
                          .drop(columns='timedelta_col'))


@pytest.fixture(params=['describe_df_pandas', 'describe_df_dask', 'describe_df_koalas'])
def describe_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_same_mi_pandas():
    return pd.DataFrame({
        'ints': pd.Series([2, pd.NA, 5, 2], dtype='Int64'),
        'floats': pd.Series([1, None, 100, 1]),
        'nans': pd.Series([None, None, None, None]),
        'nat_lang': pd.Series(['this is a very long sentence inferred as a string', None, 'test', 'test']),
    })


@pytest.fixture()
def df_same_mi_dask(df_same_mi_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(df_same_mi_pandas, npartitions=2)


@pytest.fixture()
def df_same_mi_koalas(df_same_mi_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    df_same_mi_pandas['ints'] = df_same_mi_pandas['ints'].astype('float')
    df_same_mi_pandas['nans'] = df_same_mi_pandas['nans'].astype('float')
    return ks.DataFrame(df_same_mi_pandas)


@pytest.fixture(params=['df_same_mi_pandas', 'df_same_mi_dask', 'df_same_mi_koalas'])
def df_same_mi(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_mi_pandas():
    df = pd.DataFrame({
        'ints': pd.Series([1, 2, 1]),
        'bools': pd.Series([True, False, True]),
        'strs2': pd.Series(['bye', 'hi', 'bye']),
        'strs': pd.Series(['hi', 'hi', 'hi']),
        'dates': pd.Series(['2020-01-01', '2020-01-01', '1997-01-04'])
    })
    df = df.loc[df.index.repeat(4)].reset_index(drop=True)
    return df


@pytest.fixture()
def df_mi_dask(df_mi_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(df_mi_pandas, npartitions=1)


@pytest.fixture()
def df_mi_koalas(df_mi_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(df_mi_pandas)


@pytest.fixture(params=['df_mi_pandas', 'df_mi_dask', 'df_mi_koalas'])
def df_mi(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_mi_unique_pandas():
    df = pd.DataFrame({
        'unique': pd.Series(['hi', 'bye', 'hello', 'goodbye']),
        'unique_with_one_nan': pd.Series(['hi', 'bye', None, 'goodbye']),
        'unique_with_nans': pd.Series([1, None, None, 2]),
        'ints': pd.Series([1, 2, 1, 2]),
    })
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    return df


@pytest.fixture()
def df_mi_unique_dask(df_mi_unique_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(df_mi_unique_pandas, npartitions=1)


@pytest.fixture()
def df_mi_unique_koalas(df_mi_unique_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(df_mi_unique_pandas)


@pytest.fixture(params=['df_mi_unique_pandas', 'df_mi_unique_dask', 'df_mi_unique_koalas'])
def df_mi_unique(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def categorical_df_pandas():
    return pd.DataFrame({
        'ints': pd.Series([1, 2, 3, 2]),
        'categories1': pd.Series([1, 100, 1, 100, 200, 200, 200, 200, 3, 100]),
        'bools': pd.Series([True, False, True, False]),
        'categories2': pd.Series(['test', 'test', 'test2', 'test']),
        'categories3': pd.Series(['test', 'test', 'test', np.nan]),
    })


@pytest.fixture()
def categorical_df_dask(categorical_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(categorical_df_pandas, npartitions=2)


@pytest.fixture()
def categorical_df_koalas(categorical_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(categorical_df_pandas)


@pytest.fixture(params=['categorical_df_pandas', 'categorical_df_dask', 'categorical_df_koalas'])
def categorical_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def empty_df_pandas():
    return pd.DataFrame({})


@pytest.fixture()
def empty_df_dask(empty_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(empty_df_pandas, npartitions=2)


# Cannot have an empty Koalas DataFrame
@pytest.fixture(params=['empty_df_pandas', 'empty_df_dask'])
def empty_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def small_df_pandas():
    return pd.DataFrame(pd.Series([pd.to_datetime('2020-09-01')] * 4, name='sample_datetime_series').astype('object'))


@pytest.fixture()
def small_df_dask(small_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(small_df_pandas, npartitions=1)


@pytest.fixture()
def small_df_koalas(small_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(small_df_pandas)


@pytest.fixture(params=['small_df_pandas', 'small_df_dask', 'small_df_koalas'])
def small_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def latlong_df_pandas():
    return pd.DataFrame({
        'tuple_ints': pd.Series([(1, 2), (3, 4)]),
        'tuple_strings': pd.Series([('1', '2'), ('3', '4')]),
        'string_tuple': pd.Series(['(1, 2)', '(3, 4)']),
        'bracketless_string_tuple': pd.Series(['1, 2', '3, 4']),
        'list_strings': pd.Series([['1', '2'], ['3', '4']]),
        'combo_tuple_types': pd.Series(['[1, 2]', '(3, 4)']),
        'null_value': pd.Series([np.nan, (3, 4)]),
        'null_latitude': pd.Series([(np.nan, 2.0), (3.0, 4.0)]),
        'both_null': pd.Series([(np.nan, np.nan), (3.0, 4.0)]),
    })


@pytest.fixture()
def latlong_df_dask(latlong_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(latlong_df_pandas, npartitions=2)


@pytest.fixture()
def latlong_df_koalas(latlong_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(latlong_df_pandas.applymap(lambda tup: list(tup) if isinstance(tup, tuple) else tup))


@pytest.fixture(params=['latlong_df_pandas', 'latlong_df_dask', 'latlong_df_koalas'])
def latlong_df(request):
    return request.getfixturevalue(request.param)


# LatLong Fixtures for testing access to latlong values
@pytest.fixture
def pandas_latlongs():
    return [
        pd.Series([(1.0, 2.0), (3.0, 4.0)]),
        pd.Series([('1', '2'), ('3', '4')]),
        pd.Series([['1', '2'], ['3', '4']]),
        pd.Series([(1, 2), (3, 4)]),
        pd.Series([[1, 2], [3, 4]]),
        pd.Series(['(1, 2)', '(3, 4)']),
        pd.Series(['1, 2', '3, 4']),
        pd.Series(['[1, 2]', '[3, 4]'])
    ]


@pytest.fixture
def dask_latlongs(pandas_latlongs):
    return [pd_to_dask(series) for series in pandas_latlongs]


@pytest.fixture
def koalas_latlongs(pandas_latlongs):
    return [pd_to_koalas(series.apply(lambda tup: list(tup) if isinstance(tup, tuple) else tup)) for series in pandas_latlongs]


@pytest.fixture(params=['pandas_latlongs', 'dask_latlongs', 'koalas_latlongs'])
def latlongs(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def whitespace_df_pandas():
    return pd.DataFrame({
        'id': [0, 1, 2, 3, 4, 5],
        'comments': [
            '\nleading newline',
            'trailing newline\n',
            '\n',
            'newline in \n the middle',
            '    leading whitespace',
            'trailing whitespace ',
        ]
    })


@pytest.fixture()
def whitespace_df_dask(whitespace_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(whitespace_df_pandas, npartitions=2)


@pytest.fixture()
def whitespace_df_koalas(whitespace_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(whitespace_df_pandas)


@pytest.fixture(params=['whitespace_df_pandas', 'whitespace_df_dask', 'whitespace_df_koalas'])
def whitespace_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def falsy_names_df_pandas():
    return pd.DataFrame({
        0: ['a', 'b', 'c'],
        '': [1, 2, 3],
    })


@pytest.fixture()
def falsy_names_df_dask(falsy_names_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(falsy_names_df_pandas, npartitions=2)


@pytest.fixture()
def falsy_names_df_koalas(falsy_names_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(falsy_names_df_pandas.applymap(lambda tup: list(tup) if isinstance(tup, tuple) else tup))


@pytest.fixture(params=['falsy_names_df_pandas', 'falsy_names_df_dask', 'falsy_names_df_koalas'])
def falsy_names_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_column_names(sample_df_pandas):
    return sample_df_pandas.columns.to_list()


@pytest.fixture()
def sample_inferred_logical_types():
    return {'id': Integer,
            'full_name': Unknown,
            'email': EmailAddress,
            'phone_number': Unknown,
            'age': IntegerNullable,
            'signup_date': Datetime,
            'is_registered': BooleanNullable,
            'double': Double,
            'double_with_nan': Double,
            'integer': Integer,
            'nullable_integer': IntegerNullable,
            'boolean': Boolean,
            'categorical': Categorical,
            'datetime_with_NaT': Datetime,
            }


@pytest.fixture()
def sample_correct_logical_types():
    return {'id': Integer,
            'full_name': PersonFullName,
            'email': EmailAddress,
            'phone_number': PhoneNumber,
            'age': IntegerNullable,
            'signup_date': Datetime,
            'is_registered': BooleanNullable,
            'double': Double,
            'double_with_nan': Double,
            'integer': Integer,
            'nullable_integer': IntegerNullable,
            'boolean': Boolean,
            'categorical': Categorical,
            'datetime_with_NaT': Datetime,
            }


@pytest.fixture()
def serialize_df_pandas():
    df = pd.DataFrame({
        'id': [0, 1, 2],
        'cat_int': [1, 2, 1],
        'ord_int': [1, 2, 1],
        'cat_float': [1.0, 2.0, 1.0],
        'ord_float': [1.0, 2.0, 1.0],
        'cat_bool': [True, False, True],
        'ord_bool': [True, False, True],
    })
    return df


@pytest.fixture()
def serialize_df_dask(serialize_df_pandas):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(serialize_df_pandas, npartitions=2)


@pytest.fixture()
def serialize_df_koalas(serialize_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(serialize_df_pandas)


@pytest.fixture(params=['serialize_df_pandas', 'serialize_df_dask', 'serialize_df_koalas'])
def serialize_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def pandas_datetimes():
    return [
        pd.Series(['3/11/2000', '3/12/2000', '3/13/2000', '3/14/2000']),
        pd.Series(['3/11/2000', np.nan, '3/13/2000', '3/14/2000']),
    ]


@pytest.fixture()
def dask_datetimes(pandas_datetimes):
    return [pd_to_dask(series) for series in pandas_datetimes]


@pytest.fixture()
def koalas_datetimes(pandas_datetimes):
    return [pd_to_koalas(series) for series in pandas_datetimes]


@pytest.fixture(params=['pandas_datetimes', 'dask_datetimes', 'koalas_datetimes'])
def datetimes(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def outliers_df_pandas():
    return pd.DataFrame({
        'has_outliers': [93, 42, 37, -16, 49, 42, 36, 57, 60, 23],
        'no_outliers': [60, 42, 37, 23, 49, 42, 36, 57, 60, 23.0],
        'non_numeric': ['a'] * 10,
        'has_outliers_with_nans': [None, 42, 37, -16, 49, 93, 36, 57, 60, 23],
        'nans': pd.Series([None] * 10, dtype='float64'),
    })


@pytest.fixture()
def outliers_df_dask(outliers_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason='Dask not installed, skipping')
    return dd.from_pandas(outliers_df_pandas, npartitions=2)


@pytest.fixture()
def outliers_df_koalas(outliers_df_pandas):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(outliers_df_pandas)


@pytest.fixture(params=['outliers_df_pandas', 'outliers_df_dask', 'outliers_df_koalas'])
def outliers_df(request):
    return request.getfixturevalue(request.param)



from timeit import default_timer as timer

import numpy as np
import pandas as pd
from sklearn.metrics.cluster import normalized_mutual_info_score

from woodwork.accessor_utils import _is_dask_dataframe, _is_koalas_dataframe
from woodwork.logical_types import (
    Datetime,
    Double,
    Integer,
    IntegerNullable,
    LatLong
)
from woodwork.utils import _update_progress, get_valid_mi_types, import_or_none

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def _get_describe_dict(dataframe, include=None, callback=None,
                       extra_stats=False, bins=10, top_x=10, recent_x=10):
    """Calculates statistics for data contained in a DataFrame using Woodwork typing information.

    Args:
        dataframe (pd.DataFrame): DataFrame to be described with Woodwork typing information initialized
        include (list[str or LogicalType], optional): filter for what columns to include in the
            statistics returned. Can be a list of column names, semantic tags, logical types, or a list
            combining any of the three. It follows the most broad specification. Favors logical types
            then semantic tag then column name. If no matching columns are found, an empty DataFrame
            will be returned.
        callback (callable, optional): function to be called with incremental updates. Has the following parameters:

            - update (int): change in progress since last call
            - progress (int): the progress so far in the calculations
            - total (int): the total number of calculations to do
            - unit (str): unit of measurement for progress/total
            - time_elapsed (float): total time in seconds elapsed since start of call

        extra_stats (bool): If True, will calculate a histogram for numeric columns, top values
            for categorical columns and value counts for the most recent values in datetime columns. Will also
            calculate value counts within the range of values present for integer columns if the range of
            values present is less than or equal to than the number of bins used to compute the histogram.
            Output can be controlled by bins, top_x and recent_x parameters.
        bins (int): Number of bins to use when calculating histogram for numeric columns. Defaults to 10.
            Will be ignored unless extra_stats=True.
        top_x (int): Number of items to return when getting the most frequently occurring values for categorical
            columns. Defaults to 10. Will be ignored unless extra_stats=True.
        recent_x (int): Number of values to return when calculating value counts for the most recent dates in
            datetime columns. Defaults to 10. Will be ignored unless extra_stats=True.

    Returns:
        dict[str -> dict]: A dictionary with a key for each column in the data or for each column
        matching the logical types, semantic tags or column names specified in ``include``, paired
        with a value containing a dictionary containing relevant statistics for that column.
    """
    start_time = timer()
    unit = 'calculations'
    agg_stats_to_calculate = {
        'category': ["count", "nunique"],
        'numeric': ["count", "max", "min", "nunique", "mean", "std"],
        Datetime: ["count", "max", "min", "nunique", "mean"],
    }
    if include is not None:
        filtered_cols = dataframe.ww._filter_cols(include, col_names=True)
        cols_to_include = [(k, v) for k, v in dataframe.ww.columns.items() if k in filtered_cols]
    else:
        cols_to_include = dataframe.ww.columns.items()

    results = {}

    if _is_dask_dataframe(dataframe):
        df = dataframe.compute()
    elif _is_koalas_dataframe(dataframe):
        df = dataframe.to_pandas()

        # Any LatLong columns will be using lists, which we must convert
        # back to tuples so we can calculate the mode, which requires hashable values
        latlong_columns = [col_name for col_name, col in dataframe.ww.columns.items() if type(col.logical_type) == LatLong]
        df[latlong_columns] = df[latlong_columns].applymap(lambda latlong: tuple(latlong) if latlong else latlong)
    else:
        df = dataframe

    # Setup for progress callback and make initial call
    # Assume 1 unit for general preprocessing, plus main loop over column
    total_loops = 1 + len(cols_to_include)
    current_progress = _update_progress(start_time, timer(), 1, 0, total_loops, unit, callback)

    for column_name, column in cols_to_include:
        if 'index' in column.semantic_tags:
            current_progress = _update_progress(start_time, timer(), 1, current_progress, total_loops, unit, callback)
            continue
        values = {}
        logical_type = column.logical_type
        semantic_tags = column.semantic_tags
        series = df[column_name]

        # Calculate Aggregation Stats
        if column.is_categorical:
            agg_stats = agg_stats_to_calculate['category']
        elif column.is_numeric:
            agg_stats = agg_stats_to_calculate['numeric']
        elif column.is_datetime:
            agg_stats = agg_stats_to_calculate[Datetime]
        else:
            agg_stats = ["count"]
        values = series.agg(agg_stats).to_dict()

        # Calculate other specific stats based on logical type or semantic tags
        if column.is_boolean:
            values["num_false"] = series.value_counts().get(False, 0)
            values["num_true"] = series.value_counts().get(True, 0)
        elif column.is_numeric:
            float_series = series.astype('float64')  # workaround for https://github.com/pandas-dev/pandas/issues/42626
            quant_values = float_series.quantile([0.25, 0.5, 0.75]).tolist()
            values["first_quartile"] = quant_values[0]
            values["second_quartile"] = quant_values[1]
            values["third_quartile"] = quant_values[2]

        mode = _get_mode(series)
        # The format of the mode should match its format in the DataFrame
        if _is_koalas_dataframe(dataframe) and series.name in latlong_columns:
            mode = list(mode)

        values["nan_count"] = series.isna().sum()
        values["mode"] = mode
        values["physical_type"] = series.dtype
        values["logical_type"] = logical_type
        values["semantic_tags"] = semantic_tags

        # Calculate extra detailed stats, if requested
        if extra_stats:
            if column.is_numeric:
                values["histogram"] = _get_histogram_values(series, bins=bins)
                if isinstance(column.logical_type, (Integer, IntegerNullable)):
                    _range = range(int(values["min"]), int(values["max"]) + 1)
                    # Calculate top numeric values if range of values present
                    # is less than or equal number of histogram bins
                    if len(_range) <= bins:
                        values["top_values"] = _get_numeric_value_counts_in_range(series, _range)
            elif column.is_categorical:
                values["top_values"] = _get_top_values_categorical(series, top_x)
            elif column.is_datetime:
                values["recent_values"] = _get_recent_value_counts(series, recent_x)

        results[column_name] = values
        current_progress = _update_progress(start_time, timer(), 1, current_progress, total_loops, unit, callback)
    return results


def _get_mode(series):
    """Get the mode value for a series"""
    mode_values = series.mode()
    if len(mode_values) > 0:
        return mode_values[0]
    return None


def _replace_nans_for_mutual_info(schema, data):
    """Replace NaN values in the dataframe so that mutual information can be calculated

    Args:
        schema (woodwork.TableSchema): Woodwork typing info for the data
        data (pd.DataFrame): dataframe to use for calculating mutual information

    Returns:
        pd.DataFrame: data with nans replaced with either mean or mode

    """
    for column_name in data.columns[data.isnull().any()]:
        column = schema.columns[column_name]
        series = data[column_name]

        if column.is_numeric or column.is_datetime:
            mean = series.mean()
            if isinstance(mean, float) and not mean.is_integer() and not type(column.logical_type) == Double:
                data[column_name] = series.astype('float')
            data[column_name] = data[column_name].fillna(mean)
        elif column.is_categorical or column.is_boolean:
            mode = _get_mode(series)
            data[column_name] = series.fillna(mode)
    return data


def _make_categorical_for_mutual_info(schema, data, num_bins):
    """Transforms dataframe columns into numeric categories so that
    mutual information can be calculated

    Args:
        schema (woodwork.TableSchema): Woodwork typing info for the data
        data (pd.DataFrame): dataframe to use for calculating mutual information
        num_bins (int): Determines number of bins to use for converting
            numeric features into categorical.

    Returns:
        pd.DataFrame: data with values transformed and binned into numeric categorical values
    """

    for col_name in data.columns:
        column = schema.columns[col_name]
        if column.is_numeric:
            # bin numeric features to make categories
            data[col_name] = pd.qcut(data[col_name], num_bins, duplicates="drop")
        # Convert Datetimes to total seconds - an integer - and bin
        if column.is_datetime:
            data[col_name] = pd.qcut(data[col_name].view('int64'), num_bins, duplicates="drop")
        # convert categories to integers
        new_col = data[col_name]
        if str(new_col.dtype) != 'category':
            new_col = new_col.astype('category')
        data[col_name] = new_col.cat.codes
    return data


def _get_mutual_information_dict(dataframe, num_bins=10, nrows=None, include_index=False, callback=None):
    """Calculates mutual information between all pairs of columns in the DataFrame that
    support mutual information. Logical Types that support mutual information are as
    follows:  Boolean, Categorical, CountryCode, Datetime, Double, Integer, Ordinal,
    PostalCode, and SubRegionCode

    Args:
        dataframe (pd.DataFrame): Data containing Woodwork typing information
            from which to calculate mutual information.
        num_bins (int): Determines number of bins to use for converting
            numeric features into categorical.
        nrows (int): The number of rows to sample for when determining mutual info.
            If specified, samples the desired number of rows from the data.
            Defaults to using all rows.
        include_index (bool): If True, the column specified as the index will be
            included as long as its LogicalType is valid for mutual information calculations.
            If False, the index column will not have mutual information calculated for it.
            Defaults to False.
        callback (callable, optional): function to be called with incremental updates. Has the following parameters:

            - update (int): change in progress since last call
            - progress (int): the progress so far in the calculations
            - total (int): the total number of calculations to do
            - unit (str): unit of measurement for progress/total
            - time_elapsed (float): total time in seconds elapsed since start of call

    Returns:
        list(dict): A list containing dictionaries that have keys `column_1`,
        `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
        Mutual information values are between 0 (no mutual information) and 1
        (perfect dependency).
        """
    start_time = timer()
    unit = 'calculations'
    valid_types = get_valid_mi_types()
    valid_columns = [col_name for col_name, col in dataframe.ww.columns.items() if type(col.logical_type) in valid_types]

    if not include_index and dataframe.ww.index is not None:
        valid_columns.remove(dataframe.ww.index)

    data = dataframe.loc[:, valid_columns]
    if _is_dask_dataframe(data):
        data = data.compute()
    if _is_koalas_dataframe(dataframe):
        data = data.to_pandas()

    # cut off data if necessary
    if nrows is not None and nrows < data.shape[0]:
        data = data.sample(nrows)

    # remove fully null columns
    not_null_cols = data.columns[data.notnull().any()]
    if set(not_null_cols) != set(valid_columns):
        data = data.loc[:, not_null_cols]

    # Setup for progress callback and make initial call
    # Assume 1 unit for preprocessing, n for replace nans, n for make categorical and (n*n+n)/2 for main calculation loop
    n = len(data.columns)
    total_loops = 1 + 2 * n + (n * n + n) / 2
    current_progress = _update_progress(start_time, timer(), 1, 0, total_loops, unit, callback)

    data = _replace_nans_for_mutual_info(dataframe.ww.schema, data)
    current_progress = _update_progress(start_time, timer(), n, current_progress, total_loops, unit, callback)

    data = _make_categorical_for_mutual_info(dataframe.ww.schema, data, num_bins)
    current_progress = _update_progress(start_time, timer(), n, current_progress, total_loops, unit, callback)

    # calculate mutual info for all pairs of columns
    mutual_info = []
    col_names = data.columns.to_list()
    for i, a_col in enumerate(col_names):
        for j in range(i, len(col_names)):
            b_col = col_names[j]
            if a_col == b_col:
                # Ignore because the mutual info for a column with itself will always be 1
                current_progress = _update_progress(start_time, timer(), 1, current_progress, total_loops, unit, callback)
                continue
            else:
                mi_score = normalized_mutual_info_score(data[a_col], data[b_col])
                mutual_info.append(
                    {"column_1": a_col, "column_2": b_col, "mutual_info": mi_score}
                )
                current_progress = _update_progress(start_time, timer(), 1, current_progress, total_loops, unit, callback)
    mutual_info.sort(key=lambda mi: mi['mutual_info'], reverse=True)

    return mutual_info


def _get_value_counts(dataframe, ascending=False, top_n=10, dropna=False):
    """Returns a list of dictionaries with counts for the most frequent values in each column (only
        for columns with `category` as a standard tag).


    Args:
        dataframe (pd.DataFrame, dd.DataFrame, ks.DataFrame): Data from which to count values.
        ascending (bool): Defines whether each list of values should be sorted most frequent
            to least frequent value (False), or least frequent to most frequent value (True).
            Defaults to False.

        top_n (int): the number of top values to retrieve. Defaults to 10.

        dropna (bool): determines whether to remove NaN values when finding frequency. Defaults
            to False.

    Returns:
        list(dict): a list of dictionaries for each categorical column with keys `count`
        and `value`.
    """
    val_counts = {}
    valid_cols = [col for col, column in dataframe.ww.columns.items() if column.is_categorical]
    data = dataframe[valid_cols]
    is_ks = False
    if _is_dask_dataframe(data):
        data = data.compute()
    if _is_koalas_dataframe(data):
        data = data.to_pandas()
        is_ks = True

    for col in valid_cols:
        if dropna and is_ks:
            # Koalas categorical columns will have missing values replaced with the string 'None'
            # Replace them with np.nan so dropna work
            datacol = data[col].replace(to_replace='None', value=np.nan)
        else:
            datacol = data[col]
        frequencies = datacol.value_counts(ascending=ascending, dropna=dropna)
        df = frequencies[:top_n].reset_index()
        df.columns = ["value", "count"]
        values = list(df.to_dict(orient="index").values())
        val_counts[col] = values
    return val_counts


def _get_box_plot_info_for_column(series, quantiles=None):
    """Gets the information necessary to create a box and whisker plot with outliers for a numeric column
        using the IQR method.

    Args:
        series (Series): Data for which the box plot and outlier information will be gathered.
            Will be used to calculate quantiles if none are provided.
        quantiles (dict[float -> float], optional): A dictionary containing the quantiles for the data
            where the key indicates the quantile, and the value is the quantile's value for the data.

    Note:
        The minimum quantiles necessary for outlier detection using the IQR method are the
        first quartile (0.25) and third quartile (0.75). If these keys are missing from the quantiles dictionary,
        the following quantiles will be calculated: {0.0, 0.25, 0.5, 0.75, 1.0}, which correspond to
        {min, first quantile, median, third quantile, max}.

    Returns:
        (dict[str -> float,list[number]]): Returns a dictionary containing box plot information for the Series.
            The following elements will be found in the dictionary:

            - low_bound (float): the lower bound below which outliers lay - to be used as a whisker
            - high_bound (float): the high bound above which outliers lay - to be used as a whisker
            - quantiles (list[float]): the quantiles used to determine the bounds.
                If quantiles were passed in, will contain all quantiles passed in. Otherwise, contains the five
                quantiles {0.0, 0.25, 0.5, 0.75, 1.0}.
            - low_values (list[float, int]): the values of the lower outliers
            - high_values (list[float, int]): the values of the upper outliers
            - low_indices (list[int]): the corresponding index values for each of the lower outliers
            - high_indices (list[int]): the corresponding index values for each of the upper outliers
    """
    if not series.ww._schema.is_numeric:
        raise TypeError('Cannot calculate box plot statistics for non-numeric column')

    if quantiles and not isinstance(quantiles, dict):
        raise TypeError('quantiles must be a dictionary.')

    if dd and isinstance(series, dd.Series):
        series = series.compute()
    if ks and isinstance(series, ks.Series):
        series = series.to_pandas()

    # remove null values from the data
    series = series.dropna()

    # An empty or fully null Series has no outliers, bounds, or quantiles
    if series.shape[0] == 0:
        return {
            'low_bound': np.nan,
            'high_bound': np.nan,
            'quantiles': {0.0: np.nan, 0.25: np.nan, 0.5: np.nan, 0.75: np.nan, 1.0: np.nan},
            "low_values": [],
            "high_values": [],
            "low_indices": [],
            "high_indices": []
        }

    # calculate the outlier bounds using IQR
    if quantiles is None:
        quantiles = series.quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict()
    elif 0.25 not in quantiles or 0.75 not in quantiles:
        raise ValueError("Input quantiles do not contain the minimum necessary quantiles for outlier calculation: "
                         "0.25 (the first quartile) and 0.75 (the third quartile).")

    q1 = quantiles[0.25]
    q3 = quantiles[0.75]

    iqr = q3 - q1

    low_bound = q1 - (iqr * 1.5)
    high_bound = q3 + (iqr * 1.5)

    # identify outliers in the series
    min = quantiles.get(0.0)
    max = quantiles.get(1.0)
    if min is not None and max is not None and low_bound <= min and high_bound >= max:
        outliers_dict = {
            "low_values": [],
            "high_values": [],
            "low_indices": [],
            "high_indices": []
        }
    else:
        low_series = series[series < low_bound]
        high_series = series[series > high_bound]

        outliers_dict = {
            "low_values": low_series.tolist(),
            "high_values": high_series.tolist(),
            "low_indices": low_series.index.tolist(),
            "high_indices": high_series.index.tolist()
        }

    return {'low_bound': low_bound,
            'high_bound': high_bound,
            'quantiles': quantiles,
            **outliers_dict}


def _get_numeric_value_counts_in_range(series, _range):
    """Count the number of occurrences of integers present in a series with values defined
    by a range of integers. Null values will be ignored.

    Args:
        series (pd.Series): data from which to determine the number of occurrences of values
        _range (type(range)): sequence of integers defining the values for which counts should be made

    Returns:
        value_counts (list(dict)): a list of dictionaries with keys `value` and
            `count`. Output is sorted in descending order based on the value counts.
    """
    frequencies = series.value_counts(dropna=True)
    value_counts = [{"value": i, "count": frequencies[i] if i in frequencies else 0} for i in _range]
    return sorted(value_counts, key=lambda i: (-i["count"], i["value"]))


def _get_top_values_categorical(series, num_x):
    """Get the most frequent values in a pandas Series. Will exclude null values.

    Args:
        column (pd.Series): data to use find most frequent values
        num_x (int): the number of top values to retrieve

    Returns:
        top_list (list(dict)): a list of dictionary with keys `value` and `count`.
            Output is sorted in descending order based on the value counts.
    """
    frequencies = series.value_counts(dropna=True)
    df = frequencies.head(num_x).reset_index()
    df.columns = ["value", "count"]
    df = df.sort_values(["count", "value"], ascending=[False, True])
    value_counts = list(df.to_dict(orient="index").values())
    return value_counts


def _get_recent_value_counts(column, num_x):
    """Get the the number of occurrences of the x most recent values in a datetime column.

    Args:
        column (pd.Series): data to use find value counts
        num_x (int): the number of values to retrieve

    Returns:
        value_counts (list(dict)): a list of dictionary with keys `value` and
            `count`. Output is sorted in descending order based on the value counts.
    """
    datetimes = getattr(column.dt, "date")
    frequencies = datetimes.value_counts(dropna=False)
    values = frequencies.sort_index(ascending=False)[:num_x]
    df = values.reset_index()
    df.columns = ["value", "count"]
    df = df.sort_values(["count", "value"], ascending=[False, True])
    value_counts = list(df.to_dict(orient="index").values())
    return value_counts


def _get_histogram_values(series, bins=10):
    """Get the histogram for a given numeric column.

    Args:
        series (pd.Series): data to use for histogram
        bins (int): the number of bins to use for the histogram

    Returns:
        histogram (list(dict)): a list of dictionary with keys `bins` and
            `frequency`
    """
    values = (
        pd.cut(series, bins=bins, duplicates='drop').value_counts().sort_index()
    )
    df = values.reset_index()
    df.columns = ["bins", "frequency"]
    return list(df.to_dict(orient="index").values())



import re

import pytest
from mock import patch

from woodwork.exceptions import ColumnNotPresentError
from woodwork.logical_types import (
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
    Ordinal
)
from woodwork.table_schema import (
    TableSchema,
    _check_column_descriptions,
    _check_column_metadata,
    _check_column_names,
    _check_column_origins,
    _check_index,
    _check_logical_types,
    _check_semantic_tags,
    _check_table_metadata,
    _check_time_index,
    _check_use_standard_tags,
    _validate_params
)


def test_validate_params_errors(sample_column_names):
    error_message = 'Table name must be a string'
    with pytest.raises(TypeError, match=error_message):
        _validate_params(column_names=sample_column_names,
                         name=1,
                         index=None,
                         time_index=None,
                         logical_types=None,
                         table_metadata=None,
                         column_metadata=None,
                         semantic_tags=None,
                         column_descriptions=None,
                         column_origins=None,
                         use_standard_tags=False)


def test_check_index_errors(sample_column_names):
    error_message = 'Specified index column `foo` not found in TableSchema.'
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_index(column_names=sample_column_names, index='foo')


def test_check_time_index_errors(sample_column_names):
    error_message = 'Specified time index column `foo` not found in TableSchema'
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_time_index(column_names=sample_column_names, time_index='foo', logical_type=Integer)

    error_msg = 'Time index column must be a Datetime or numeric column'
    with pytest.raises(TypeError, match=error_msg):
        _check_time_index(column_names=sample_column_names, time_index='full_name', logical_type=NaturalLanguage)


def test_check_column_names(sample_column_names):
    error_message = 'Column names must be a list or set'
    with pytest.raises(TypeError, match=error_message):
        _check_column_names(column_names=int)

    sample_column_names.append('id')
    with pytest.raises(IndexError, match='TableSchema cannot contain duplicate columns names'):
        _check_column_names(sample_column_names)


def test_check_logical_types_errors(sample_column_names):
    error_message = 'logical_types must be a dictionary'
    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(sample_column_names, logical_types='type')

    bad_logical_types_keys = {
        'full_name': None,
        'age': None,
        'birthday': None,
        'occupation': None,
    }
    error_message = re.escape("logical_types contains columns that are not present in TableSchema: ['birthday', 'occupation']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys)

    bad_logical_types_keys = {
        'id': None,
        'full_name': None,
        'email': None,
        'phone_number': None,
        'age': None,
    }
    error_message = re.escape("logical_types is missing columns that are present in TableSchema: "
                              "['boolean', 'categorical', 'datetime_with_NaT', 'double', 'double_with_nan', "
                              "'integer', 'is_registered', 'nullable_integer', 'signup_date']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys)

    bad_logical_types_keys = {'email': 1}
    error_message = ("Logical Types must be of the LogicalType class "
                     "and registered in Woodwork's type system. "
                     "1 does not meet that criteria.")

    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys, require_all_cols=False)

    bad_logical_types_keys = {
        'email': 'NaturalLanguage',
    }
    error_message = ("Logical Types must be of the LogicalType class "
                     "and registered in Woodwork's type system. "
                     "NaturalLanguage does not meet that criteria.")

    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys, require_all_cols=False)


def test_check_semantic_tags_errors(sample_column_names):
    error_message = 'semantic_tags must be a dictionary'
    with pytest.raises(TypeError, match=error_message):
        _check_semantic_tags(sample_column_names, semantic_tags='type')

    bad_semantic_tags_keys = {
        'full_name': None,
        'age': None,
        'birthday': None,
        'occupation': None,
    }
    error_message = re.escape("semantic_tags contains columns that are not present in TableSchema: ['birthday', 'occupation']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_semantic_tags(sample_column_names, bad_semantic_tags_keys)

    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _check_semantic_tags(sample_column_names, {'id': 1})


def test_check_table_metadata_errors():
    error_message = 'Table metadata must be a dictionary.'
    with pytest.raises(TypeError, match=error_message):
        _check_table_metadata('test')


def test_check_column_metadata_errors(sample_column_names):
    error_message = 'Column metadata must be a dictionary.'
    with pytest.raises(TypeError, match=error_message):
        _check_column_metadata(sample_column_names, column_metadata='test')

    column_metadata = {
        'invalid_col': {'description': 'not a valid column'}
    }
    err_msg = re.escape("column_metadata contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_metadata(sample_column_names, column_metadata=column_metadata)


def test_check_column_description_errors(sample_column_names):
    error_message = 'column_descriptions must be a dictionary'
    with pytest.raises(TypeError, match=error_message):
        _check_column_descriptions(sample_column_names, column_descriptions='test')

    column_descriptions = {
        'invalid_col': 'a description'
    }
    err_msg = re.escape("column_descriptions contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_descriptions(sample_column_names, column_descriptions=column_descriptions)


def test_check_column_origin_errors(sample_column_names):
    error_message = 'column_origins must be a dictionary or a string'
    with pytest.raises(TypeError, match=error_message):
        _check_column_origins(sample_column_names, column_origins=123)

    column_origins = {
        'invalid_col': 'base'
    }
    err_msg = re.escape("column_origins contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_origins(sample_column_names, column_origins=column_origins)


def test_check_use_standard_tags_errors(sample_column_names):
    error_message = 'use_standard_tags must be a dictionary or a boolean'
    with pytest.raises(TypeError, match=error_message):
        _check_use_standard_tags(sample_column_names, use_standard_tags=1)

    error_message = re.escape("use_standard_tags contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_use_standard_tags(sample_column_names, use_standard_tags={'invalid_col': True})

    error_message = "use_standard_tags for column id must be a boolean"
    with pytest.raises(TypeError, match=error_message):
        _check_use_standard_tags(sample_column_names, use_standard_tags={'id': 1})


def test_schema_init(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    assert schema.name is None
    assert schema.index is None
    assert schema.time_index is None

    assert set(schema.columns.keys()) == set(sample_column_names)


def test_schema_init_with_name(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         name='schema')

    assert schema.name == 'schema'
    assert schema.index is None
    assert schema.time_index is None


def test_schema_init_with_name_and_indices(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         name='schema',
                         index='id',
                         time_index='signup_date')

    assert schema.name == 'schema'
    assert schema.index == 'id'
    assert schema.time_index == 'signup_date'
    assert isinstance(schema.columns[schema.time_index].logical_type, Datetime)


def test_schema_with_numeric_time_index(sample_column_names, sample_inferred_logical_types):
    # Set a numeric time index on init
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{'signup_date': Integer}},
        time_index='signup_date',
        use_standard_tags=True
    )
    date_col = schema.columns['signup_date']
    assert schema.time_index == 'signup_date'
    assert isinstance(date_col.logical_type, Integer)
    assert date_col.semantic_tags == {'time_index', 'numeric'}

    # Specify logical type for time index on init
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{'signup_date': Double}},
        time_index='signup_date',
        use_standard_tags=True
    )
    date_col = schema.columns['signup_date']
    assert schema.time_index == 'signup_date'
    assert isinstance(date_col.logical_type, Double)
    assert date_col.semantic_tags == {'time_index', 'numeric'}


def test_schema_init_with_logical_type_classes(sample_column_names, sample_correct_logical_types):
    schema = TableSchema(sample_column_names, logical_types=sample_correct_logical_types,
                         name='schema')
    instantiated_ltypes = {name: ltype() for name, ltype in sample_correct_logical_types.items()}
    assert schema.logical_types == instantiated_ltypes


def test_raises_error_setting_index_tag_directly(sample_column_names, sample_inferred_logical_types):
    error_msg = re.escape("Cannot add 'index' tag directly for column id. To set a column as the index, "
                          "use DataFrame.ww.set_index() instead.")
    with pytest.raises(ValueError, match=error_msg):
        semantic_tags = {'id': 'index'}
        TableSchema(sample_column_names, sample_inferred_logical_types,
                    name='schema',
                    semantic_tags=semantic_tags,
                    use_standard_tags=False)


def test_raises_error_setting_time_index_tag_directly(sample_column_names, sample_inferred_logical_types):
    error_msg = re.escape("Cannot add 'time_index' tag directly for column signup_date. To set a column as the time index, "
                          "use DataFrame.ww.set_time_index() instead.")
    with pytest.raises(ValueError, match=error_msg):
        semantic_tags = {'signup_date': 'time_index'}
        TableSchema(sample_column_names, sample_inferred_logical_types,
                    name='schema',
                    semantic_tags=semantic_tags,
                    use_standard_tags=False)


def test_schema_init_with_semantic_tags(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {'id': 'custom_tag'}
    schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                         name='schema',
                         semantic_tags=semantic_tags,
                         use_standard_tags=False)

    id_semantic_tags = schema.columns['id'].semantic_tags
    assert isinstance(id_semantic_tags, set)
    assert len(id_semantic_tags) == 1
    assert 'custom_tag' in id_semantic_tags


def test_schema_adds_standard_semantic_tags(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{'id': Categorical}},
        use_standard_tags=True,
        name='schema'
    )

    assert schema.semantic_tags['id'] == {'category'}
    assert schema.semantic_tags['age'] == {'numeric'}

    schema = TableSchema(sample_column_names, logical_types={**sample_inferred_logical_types, **{'id': Categorical}},
                         name='schema',
                         use_standard_tags=False)

    assert schema.semantic_tags['id'] == set()
    assert schema.semantic_tags['age'] == set()


def test_semantic_tags_during_init(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'phone_number': ['tag3'],
        'signup_date': ['secondary_time_index'],
        'age': ['numeric', 'age']
    }
    expected_types = {
        'full_name': {'tag1'},
        'email': {'tag2'},
        'phone_number': {'tag3'},
        'signup_date': {'secondary_time_index'},
        'age': {'numeric', 'age'}
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags=semantic_tags)
    assert schema.columns['full_name'].semantic_tags == expected_types['full_name']
    assert schema.columns['email'].semantic_tags == expected_types['email']
    assert schema.columns['phone_number'].semantic_tags == expected_types['phone_number']
    assert schema.columns['signup_date'].semantic_tags == expected_types['signup_date']
    assert schema.columns['age'].semantic_tags == expected_types['age']


def test_semantic_tag_errors(sample_column_names, sample_inferred_logical_types):
    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags={'id': int})

    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags={'id': {'index': {}, 'time_index': {}}})

    error_message = "semantic_tags for id must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(sample_column_names, sample_inferred_logical_types, semantic_tags={'id': ['index', 1]})


def test_index_replacing_standard_tags(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, use_standard_tags=True)
    assert schema.columns['id'].semantic_tags == {'numeric'}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types, index='id')
    assert schema.columns['id'].semantic_tags == {'index'}


def test_schema_init_with_col_descriptions(sample_column_names, sample_inferred_logical_types):
    descriptions = {
        'age': 'age of the user',
        'signup_date': 'date of account creation'
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, column_descriptions=descriptions)
    for name, column in schema.columns.items():
        assert column.description == descriptions.get(name)


def test_schema_col_descriptions_errors(sample_column_names, sample_inferred_logical_types):
    err_msg = 'column_descriptions must be a dictionary'
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(sample_column_names, sample_inferred_logical_types, column_descriptions=34)

    descriptions = {
        'invalid_col': 'not a valid column',
        'signup_date': 'date of account creation'
    }
    err_msg = re.escape("column_descriptions contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        TableSchema(sample_column_names, sample_inferred_logical_types, column_descriptions=descriptions)


def test_schema_init_with_col_origins(sample_column_names, sample_inferred_logical_types):
    origins = {
        'age': 'base',
        'signup_date': 'engineered'
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, column_origins=origins)
    for name, column in schema.columns.items():
        assert column.origin == origins.get(name)

    schema_single_origin = TableSchema(sample_column_names, sample_inferred_logical_types, column_origins='base')
    for name, column in schema_single_origin.columns.items():
        assert column.origin == 'base'


def test_schema_col_origins_errors(sample_column_names, sample_inferred_logical_types):
    err_msg = 'column_origins must be a dictionary or a string'
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(sample_column_names, sample_inferred_logical_types, column_origins=34)

    origins = {
        'invalid_col': 'not a valid column',
        'signup_date': 'base'
    }
    err_msg = re.escape("column_origins contains columns that are not present in TableSchema: ['invalid_col']")
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        TableSchema(sample_column_names, sample_inferred_logical_types, column_origins=origins)

    origins = {
        'signup_date': 1
    }
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(sample_column_names, sample_inferred_logical_types, column_origins=origins)


def test_schema_init_with_column_metadata(sample_column_names, sample_inferred_logical_types):
    column_metadata = {
        'age': {'interesting_values': [33]},
        'signup_date': {'description': 'date of account creation'}
    }
    schema = TableSchema(sample_column_names, sample_inferred_logical_types, column_metadata=column_metadata)
    for name, column in schema.columns.items():
        assert column.metadata == (column_metadata.get(name) or {})


@patch("woodwork.table_schema._validate_not_setting_index_tags")
@patch("woodwork.table_schema._check_time_index")
@patch("woodwork.table_schema._check_index")
@patch("woodwork.table_schema._validate_params")
def test_validation_methods_called(mock_validate_params, mock_check_index,
                                   mock_check_time_index, mock_validate_not_setting_index,
                                   sample_column_names, sample_inferred_logical_types):
    assert not mock_validate_params.called
    assert not mock_check_index.called
    assert not mock_check_time_index.called
    assert not mock_validate_not_setting_index.called

    not_validated_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                       index='id', time_index='signup_date', validate=False)
    assert not mock_validate_params.called
    assert not mock_check_index.called
    assert not mock_check_time_index.called
    assert not mock_validate_not_setting_index.called

    validated_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                   index='id', time_index='signup_date', validate=True)
    assert mock_validate_params.called
    assert mock_check_index.called
    assert mock_check_time_index.called
    assert mock_validate_not_setting_index.called

    assert validated_schema == not_validated_schema


def test_use_standard_tags_from_bool(sample_column_names, sample_inferred_logical_types):
    standard_tags_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                       use_standard_tags=True)
    assert set(standard_tags_schema.columns.keys()) == set(standard_tags_schema.use_standard_tags.keys())
    assert all([*standard_tags_schema.use_standard_tags.values()])
    assert standard_tags_schema.semantic_tags['id'] == {'numeric'}

    no_standard_tags_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                          use_standard_tags=False)
    assert set(no_standard_tags_schema.columns.keys()) == set(no_standard_tags_schema.use_standard_tags.keys())
    assert not any([*no_standard_tags_schema.use_standard_tags.values()])
    assert no_standard_tags_schema.semantic_tags['id'] == set()


def test_use_standard_tags_from_dict(sample_column_names, sample_inferred_logical_types):
    default_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                 use_standard_tags={col_name: False for col_name in sample_column_names})
    assert default_schema.use_standard_tags == {col_name: False for col_name in sample_column_names}

    use_standard_tags = {'id': True,
                         'full_name': False,
                         'email': True,
                         'phone_number': True,
                         'age': False,
                         'signup_date': True,
                         'is_registered': False,
                         'double': False,
                         'double_with_nan': False,
                         'integer': False,
                         'nullable_integer': False,
                         'boolean': False,
                         'categorical': False,
                         'datetime_with_NaT': False}
    full_dict_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                   use_standard_tags=use_standard_tags)
    assert full_dict_schema.use_standard_tags == use_standard_tags

    partial_dict_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                      use_standard_tags={'id': True,
                                                         'email': True,
                                                         'phone_number': True,
                                                         'signup_date': True})
    assert full_dict_schema.use_standard_tags == partial_dict_schema.use_standard_tags
    assert full_dict_schema == partial_dict_schema

    partial_dict_default_schema = TableSchema(sample_column_names, sample_inferred_logical_types,
                                              use_standard_tags={'id': False,
                                                                 'email': False,
                                                                 'phone_number': False,
                                                                 'signup_date': False})
    assert default_schema.use_standard_tags == partial_dict_default_schema.use_standard_tags
    assert default_schema == partial_dict_default_schema


def test_ordinal_without_init():
    schema = TableSchema(column_names=['ordinal_col'], logical_types={'ordinal_col': Ordinal})
    assert isinstance(schema.logical_types['ordinal_col'], Ordinal)
    assert schema.logical_types['ordinal_col'].order is None




import numpy as np
import pandas as pd
import pytest

from woodwork.logical_types import (
    Categorical,
    CountryCode,
    Double,
    Integer,
    Unknown
)
from woodwork.type_sys.inference_functions import (
    categorical_func,
    double_func,
    integer_func
)
from woodwork.type_sys.type_system import TypeSystem


def pd_to_dask(series):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(series, npartitions=2)


def pd_to_koalas(series):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(series)


# Integer Inference Fixtures
@pytest.fixture
def pandas_integers():
    return [
        pd.Series(4 * [-1, 2, 1, 7]),
        pd.Series(4 * [-1, 0, 5, 3]),
    ]


@pytest.fixture
def dask_integers(pandas_integers):
    return [pd_to_dask(series) for series in pandas_integers]


@pytest.fixture
def koalas_integers(pandas_integers):
    return [pd_to_koalas(series) for series in pandas_integers]


@pytest.fixture(params=['pandas_integers', 'dask_integers', 'koalas_integers'])
def integers(request):
    return request.getfixturevalue(request.param)


# Double Inference Fixtures
@pytest.fixture
def pandas_doubles():
    return [
        pd.Series(4 * [-1, 2.5, 1, 7]),
        pd.Series(4 * [1.5, np.nan, 1, 3])
    ]


@pytest.fixture
def dask_doubles(pandas_doubles):
    return [pd_to_dask(series) for series in pandas_doubles]


@pytest.fixture
def koalas_doubles(pandas_doubles):
    return [pd_to_koalas(series) for series in pandas_doubles]


@pytest.fixture(params=['pandas_doubles', 'dask_doubles', 'koalas_doubles'])
def doubles(request):
    return request.getfixturevalue(request.param)


# Boolean Inference Fixtures
@pytest.fixture
def pandas_bools():
    return [
        pd.Series([True, False, True, True]),
        pd.Series([True, np.nan, True, True]),
    ]


@pytest.fixture
def dask_bools(pandas_bools):
    return [pd_to_dask(series) for series in pandas_bools]


@pytest.fixture
def koalas_bools(pandas_bools):
    return [pd_to_koalas(series) for series in pandas_bools]


@pytest.fixture(params=['pandas_bools', 'dask_bools', 'koalas_bools'])
def bools(request):
    return request.getfixturevalue(request.param)


# Datetime Inference Fixtures
@pytest.fixture
def pandas_datetimes():
    return [
        pd.Series(['2000-3-11', '2000-3-12', '2000-03-13', '2000-03-14']),
        pd.Series(['2000-3-11', np.nan, '2000-03-13', '2000-03-14']),
    ]


@pytest.fixture
def dask_datetimes(pandas_datetimes):
    return [pd_to_dask(series) for series in pandas_datetimes]


@pytest.fixture
def koalas_datetimes(pandas_datetimes):
    return [pd_to_koalas(series) for series in pandas_datetimes]


@pytest.fixture(params=['pandas_datetimes', 'dask_datetimes', 'koalas_datetimes'])
def datetimes(request):
    return request.getfixturevalue(request.param)


# Email Inference Fixtures
@pytest.fixture
def pandas_emails():
    return [
        pd.Series(['fl@alteryx.com', 'good@email.com', 'boaty@mcboatface.com', 'foo@bar.com']),
        pd.Series(['fl@alteryx.com', 'good@email.com', 'boaty@mcboatface.com', np.nan]),
    ]


@pytest.fixture
def dask_emails(pandas_emails):
    return [pd_to_dask(series) for series in pandas_emails]


@pytest.fixture
def koalas_emails(pandas_emails):
    return [pd_to_koalas(series) for series in pandas_emails]


@pytest.fixture(params=['pandas_emails', 'dask_emails', 'koalas_emails'])
def emails(request):
    return request.getfixturevalue(request.param)


# Email Inference Fixtures
@pytest.fixture
def bad_pandas_emails():
    return [
        pd.Series(['fl@alteryx.com', 'not_an_email', 'good@email.com', 'foo@bar.com']),
        pd.Series(['fl@alteryx.com', 'b☃d@email.com', 'good@email.com', np.nan]),
        pd.Series(['fl@alteryx.com', '@email.com', 'good@email.com', 'foo@bar.com']),
        pd.Series(['fl@alteryx.com', 'bad@email', 'good@email.com', np.nan]),
    ]


@pytest.fixture
def bad_dask_emails(bad_pandas_emails):
    return [pd_to_dask(series) for series in bad_pandas_emails]


@pytest.fixture
def bad_koalas_emails(bad_pandas_emails):
    return [pd_to_koalas(series) for series in bad_pandas_emails]


@pytest.fixture(params=['bad_pandas_emails', 'bad_dask_emails', 'bad_koalas_emails'])
def bad_emails(request):
    return request.getfixturevalue(request.param)


# Categorical Inference Fixtures
@pytest.fixture
def pandas_categories():
    return [
        pd.Series(10 * ['a', 'b', 'a', 'b']),
        pd.Series(10 * ['1', '2', '1', '2']),
        pd.Series(10 * ['a', np.nan, 'b', 'b']),
        pd.Series(10 * [1, 2, 1, 2])
    ]


@pytest.fixture
def dask_categories(pandas_categories):
    return [pd_to_dask(series) for series in pandas_categories]


@pytest.fixture
def koalas_categories(pandas_categories):
    return [pd_to_koalas(series) for series in pandas_categories]


@pytest.fixture(params=['pandas_categories', 'dask_categories', 'koalas_categories'])
def categories(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def pandas_categories_dtype():
    return pd.DataFrame({
        "cat": pd.Series(["a", "b", "c", "d"], dtype="category"),
        "non_cat": pd.Series(["a", "b", "c", "d"], dtype="string"),
    })


@pytest.fixture
def dask_categories_dtype(pandas_categories_dtype):
    return pd_to_dask(pandas_categories_dtype)


@pytest.fixture(params=['pandas_categories_dtype', 'dask_categories_dtype'])
def categories_dtype(request):
    # Koalas doesn't support the "category" dtype.  We just leave it out for
    # now.
    return request.getfixturevalue(request.param)


# Timedelta Inference Fixtures
@pytest.fixture
def pandas_timedeltas():
    return [
        pd.Series(pd.to_timedelta(range(4), unit='s')),
        pd.Series([pd.to_timedelta(1, unit='s'), np.nan])
    ]


@pytest.fixture
def dask_timedeltas(pandas_timedeltas):
    return [pd_to_dask(series) for series in pandas_timedeltas]


@pytest.fixture(params=['pandas_timedeltas', 'dask_timedeltas'])
def timedeltas(request):
    return request.getfixturevalue(request.param)


# Unknown Inference Fixtures
@pytest.fixture
def pandas_strings():
    return [
        pd.Series(['Mr. John Doe', 'Doe, Mrs. Jane', 'James Brown', 'Ms. Paige Turner']),
    ]


@pytest.fixture
def dask_strings(pandas_strings):
    return [pd_to_dask(series) for series in pandas_strings]


@pytest.fixture
def koalas_strings(pandas_strings):
    return [pd_to_koalas(series) for series in pandas_strings]


@pytest.fixture(params=['pandas_strings', 'dask_strings', 'koalas_strings'])
def strings(request):
    return request.getfixturevalue(request.param)


# pd.NA Inference Fixtures
@pytest.fixture
def pandas_pdnas():
    return [
        pd.Series(['Mr. John Doe', pd.NA, 'James Brown', 'Ms. Paige Turner']).astype('string'),
        pd.Series([1, pd.NA, 2, 3]).astype('Int64'),
        pd.Series([True, pd.NA, False, True]).astype('boolean'),
    ]


@pytest.fixture
def dask_pdnas(pandas_pdnas):
    return [pd_to_dask(series) for series in pandas_pdnas]


@pytest.fixture(params=['pandas_pdnas', 'dask_pdnas'])
def pdnas(request):
    return request.getfixturevalue(request.param)


# Null Inference Fixtures
@pytest.fixture
def pandas_nulls():
    return [
        pd.Series([pd.NA, pd.NA, pd.NA, pd.NA]),
        pd.Series([np.nan, np.nan, np.nan, np.nan]),
        pd.Series([None, None, None, None]),
        pd.Series([None, np.nan, pd.NA, None])
    ]


@pytest.fixture
def dask_nulls(pandas_nulls):
    return [pd_to_dask(series) for series in pandas_nulls]


@pytest.fixture
def koalas_nulls(pandas_nulls):
    return [pd_to_koalas(series) for series in pandas_nulls]


@pytest.fixture(params=['pandas_nulls', 'dask_nulls', 'koalas_nulls'])
def nulls(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def default_inference_functions():
    return {
        Double: double_func,
        Integer: integer_func,
        Categorical: categorical_func,
        CountryCode: None,
        Unknown: None,
    }


@pytest.fixture
def default_relationships():
    return [(Double, Integer), (Categorical, CountryCode)]


@pytest.fixture
def type_sys(default_inference_functions, default_relationships):
    return TypeSystem(inference_functions=default_inference_functions,
                      relationships=default_relationships,
                      default_type=Unknown)



import pandas as pd
import pandas.api.types as pdtypes

import woodwork as ww
from woodwork.logical_types import Categorical, Double, LogicalType


def test_register_custom_logical_type(type_sys):
    class CustomLogicalType(LogicalType):
        primary_dtype = 'string'

    def custom_func(series):
        return True

    type_sys.add_type(CustomLogicalType,
                      inference_function=custom_func,
                      parent='Categorical')
    assert CustomLogicalType in type_sys.registered_types
    assert (Categorical, CustomLogicalType) in type_sys.relationships
    assert type_sys.inference_functions[CustomLogicalType] is custom_func
    assert isinstance(type_sys.infer_logical_type(pd.Series(['a', 'b'] + 10 * ['a'])), CustomLogicalType)


def test_custom_type_with_accessor(sample_df):
    class AgesAbove20(LogicalType):
        primary_dtype = 'float64'
        standard_tags = {'age', 'numeric'}

    def ages_func(series):
        if all(series.dropna() > 20):
            return True
        return False

    ww.type_system.add_type(AgesAbove20, inference_function=ages_func, parent='IntegerNullable')
    sample_df.ww.init()
    assert isinstance(sample_df.ww['age'].ww.logical_type, AgesAbove20)
    assert sample_df.ww['age'].ww.semantic_tags == {'age', 'numeric'}
    assert sample_df['age'].dtype == 'float64'
    # Reset global type system to original settings
    ww.type_system.reset_defaults()


def test_accessor_override_default_function(sample_df):
    def new_double_func(series):
        if pdtypes.is_integer_dtype(series.dtype):
            return True
        return False

    # Update functions to cause 'age' to be recognized as Double instead fo Integer
    ww.type_system.update_inference_function('Double', inference_function=new_double_func)
    ww.type_system.update_inference_function('Integer', inference_function=None)
    sample_df.ww.init()
    assert isinstance(sample_df.ww['age'].ww.logical_type, Double)
    assert sample_df['age'].dtype == 'float64'
    # Reset global type system to original settings
    ww.type_system.reset_defaults()



import json
import os
import tarfile
import tempfile
import warnings
from itertools import zip_longest
from pathlib import Path

import pandas as pd

import woodwork as ww
from woodwork.exceptions import OutdatedSchemaWarning, UpgradeSchemaWarning
from woodwork.s3_utils import get_transport_params, use_smartopen
from woodwork.serialize import FORMATS, SCHEMA_VERSION
from woodwork.utils import _is_s3, _is_url, import_or_raise


def read_table_typing_information(path):
    """Read Woodwork typing information from disk, S3 path, or URL.

        Args:
            path (str): Location on disk, S3 path, or URL to read `woodwork_typing_info.json`.

        Returns:
            dict: Woodwork typing information dictionary
    """
    path = os.path.abspath(path)
    assert os.path.exists(path), '"{}" does not exist'.format(path)
    file = os.path.join(path, 'woodwork_typing_info.json')
    with open(file, 'r') as file:
        typing_info = json.load(file)
    typing_info['path'] = path
    return typing_info


def _typing_information_to_woodwork_table(table_typing_info, validate, **kwargs):
    """Deserialize Woodwork table from table description.

    Args:
        table_typing_info (dict) : Woodwork typing information. Likely generated using :meth:`.serialize.typing_info_to_dict`
        validate (bool): Whether parameter and data validation should occur during table initialization
        kwargs (keywords): Additional keyword arguments to pass as keywords arguments to the underlying deserialization method.

    Returns:
        DataFrame: DataFrame with Woodwork typing information initialized.
    """
    _check_schema_version(table_typing_info['schema_version'])

    path = table_typing_info['path']
    loading_info = table_typing_info['loading_info']

    file = os.path.join(path, loading_info['location'])

    load_format = loading_info['type']
    assert load_format in FORMATS

    kwargs = loading_info.get('params', {})
    table_type = loading_info.get('table_type', 'pandas')

    logical_types = {}
    semantic_tags = {}
    column_descriptions = {}
    column_origins = {}
    column_metadata = {}
    use_standard_tags = {}
    column_dtypes = {}
    for col in table_typing_info['column_typing_info']:
        col_name = col['name']

        ltype_metadata = col['logical_type']
        ltype = ww.type_system.str_to_logical_type(ltype_metadata['type'], params=ltype_metadata['parameters'])

        tags = col['semantic_tags']

        if 'index' in tags:
            tags.remove('index')
        elif 'time_index' in tags:
            tags.remove('time_index')

        logical_types[col_name] = ltype
        semantic_tags[col_name] = tags
        column_descriptions[col_name] = col['description']
        column_origins[col_name] = col['origin']
        column_metadata[col_name] = col['metadata']
        use_standard_tags[col_name] = col['use_standard_tags']

        col_type = col['physical_type']['type']
        if col_type == 'category':
            # Make sure categories are recreated properly
            cat_values = col['physical_type']['cat_values']
            cat_dtype = col['physical_type']['cat_dtype']
            if table_type == 'pandas':
                cat_object = pd.CategoricalDtype(pd.Index(cat_values, dtype=cat_dtype))
            else:
                cat_object = pd.CategoricalDtype(pd.Series(cat_values))
            col_type = cat_object
        elif table_type == 'koalas' and col_type == 'object':
            col_type = 'string'
        column_dtypes[col_name] = col_type

    if table_type == 'dask':
        DASK_ERR_MSG = (
            'Cannot load Dask DataFrame - unable to import Dask.\n\n'
            'Please install with pip or conda:\n\n'
            'python -m pip install "woodwork[dask]"\n\n'
            'conda install dask'
        )
        lib = import_or_raise('dask.dataframe', DASK_ERR_MSG)
    elif table_type == 'koalas':
        KOALAS_ERR_MSG = (
            'Cannot load Koalas DataFrame - unable to import Koalas.\n\n'
            'Please install with pip or conda:\n\n'
            'python -m pip install "woodwork[koalas]"\n\n'
            'conda install koalas\n\n'
            'conda install pyspark'
        )
        lib = import_or_raise('databricks.koalas', KOALAS_ERR_MSG)
        if 'compression' in kwargs.keys():
            kwargs['compression'] = str(kwargs['compression'])
    else:
        lib = pd

    if 'index' in kwargs.keys():
        del kwargs['index']
    if load_format == 'csv':
        dataframe = lib.read_csv(
            file,
            dtype=column_dtypes,
            **kwargs
        )
    elif load_format == 'pickle':
        dataframe = pd.read_pickle(file, **kwargs)
    elif load_format == 'parquet':
        dataframe = lib.read_parquet(file, engine=kwargs['engine'])
    elif load_format in ['arrow', 'feather']:
        dataframe = lib.read_feather(file)
    elif load_format == 'orc':
        dataframe = lib.read_orc(file)

    dataframe.ww.init(
        name=table_typing_info.get('name'),
        index=table_typing_info.get('index'),
        time_index=table_typing_info.get('time_index'),
        logical_types=logical_types,
        semantic_tags=semantic_tags,
        use_standard_tags=use_standard_tags,
        table_metadata=table_typing_info.get('table_metadata'),
        column_metadata=column_metadata,
        column_descriptions=column_descriptions,
        column_origins=column_origins,
        validate=validate)

    return dataframe


def read_woodwork_table(path, profile_name=None, validate=False, **kwargs):
    """Read Woodwork table from disk, S3 path, or URL.

        Args:
            path (str): Directory on disk, S3 path, or URL to read `woodwork_typing_info.json`.
            profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
                Set to False to use an anonymous profile.
            validate (bool, optional): Whether parameter and data validation should occur when initializing Woodwork dataframe
                during deserialization. Defaults to False. Note: If serialized data was modified outside of Woodwork and you
                are unsure of the validity of the data or typing information, `validate` should be set to True.
            kwargs (keywords): Additional keyword arguments to pass as keyword arguments to the underlying deserialization method.

        Returns:
            DataFrame: DataFrame with Woodwork typing information initialized.
    """
    if _is_url(path) or _is_s3(path):
        with tempfile.TemporaryDirectory() as tmpdir:
            file_name = Path(path).name
            file_path = os.path.join(tmpdir, file_name)
            transport_params = None

            if _is_s3(path):
                transport_params = get_transport_params(profile_name)

            use_smartopen(file_path, path, transport_params)
            with tarfile.open(str(file_path)) as tar:
                tar.extractall(path=tmpdir)

            table_typing_info = read_table_typing_information(tmpdir)
            return _typing_information_to_woodwork_table(table_typing_info, validate, **kwargs)
    else:
        table_typing_info = read_table_typing_information(path)
        return _typing_information_to_woodwork_table(table_typing_info, validate, **kwargs)


def _check_schema_version(saved_version_str):
    """Warns users if the schema used to save their data is greater than the latest
    supported schema or if it is an outdated schema that is no longer supported."""
    saved = saved_version_str.split('.')
    current = SCHEMA_VERSION.split('.')

    for c_num, s_num in zip_longest(current, saved, fillvalue=0):
        if int(c_num) > int(s_num):
            break
        elif int(c_num) < int(s_num):
            warnings.warn(UpgradeSchemaWarning().get_warning_message(saved_version_str, SCHEMA_VERSION),
                          UpgradeSchemaWarning)
            break

    # Check if saved has older major version.
    if int(current[0]) > int(saved[0]):
        warnings.warn(OutdatedSchemaWarning().get_warning_message(saved_version_str),
                      OutdatedSchemaWarning)
